import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm

# import wisdem.moorpy.MoorSolve as msolve
from wisdem.moorpy.helpers import CatenaryError, dsolve2


def catenary(XF, ZF, L, EA, W, CB=0, HF0=0, VF0=0, Tol=0.000001, nNodes=20, MaxIter=100, plots=0):
    """
    The quasi-static mooring line solver. Adapted from catenary subroutine in FAST v7 by J. Jonkman.
    Note: this version is updated Oct 7 2020 to use the dsolve solver.

    Parameters
    ----------
    XF : float
        Horizontal distance from end 1 to end 2 [m]
    ZF : float
        Vertical distance from end 1 to end 2 [m] (positive up)
    L  : float
        Unstretched length of line [m]
    EA : float
        Extensional stiffness of line [N]
    W  : float
        Weight of line in fluid per unit length [N/m]
    CB : float, optional
        If positive, coefficient of seabed static friction drag. If negative, no seabed contact and the value is the distance down from end A to the seabed in m\
            NOTE: friction (CV > 0) should only be applied when end A of the line is at an anchor, otherwise assumptions are violated.
    HF0 : float, optional
        Horizontal fairlead tension. If zero or not provided, a guess will be calculated.
    VF0 : float, optional
        Vertical fairlead tension. If zero or not provided, a guess will be calculated.

    Tol    :  float, optional
        Convergence tolerance within Newton-Raphson iteration specified as an absolute displacement error
    nNodes : int, optional
        Number of nodes to describe the line
    MaxIter:  int, optional
        Maximum number of iterations to try before resetting to default ICs and then trying again
    plots  : int, optional
        1: plot output, 0: don't


    Returns
    -------
    : tuple
        (end 1 horizontal tension, end 1 vertical tension, end 2 horizontal tension, end 2 vertical tension, info dictionary) [N] (positive up)

    """

    # make info dict to contain any additional outputs
    info = dict(error=False)

    info[
        "call"
    ] = f"catenary({XF}, {ZF}, {L}, {EA}, {W}, CB={CB}, HF0={HF0}, VF0={VF0}, Tol={Tol}, MaxIter={MaxIter}, plots=1)"

    # make some arrays if needed for plotting each node
    if plots > 0:
        s = np.linspace(
            0, L, nNodes
        )  #  Unstretched arc distance along line from anchor to each node where the line position and tension can be output (meters)
        Xs = np.zeros(nNodes)  #  Horizontal locations of each line node relative to the anchor (meters)
        Zs = np.zeros(nNodes)  #  Vertical   locations of each line node relative to the anchor (meters)
        Te = np.zeros(nNodes)  #  Effective line tensions at each node (N)

    # flip line in the solver if it is buoyant
    if W < 0:
        W = -W
        ZF = -ZF
        CB = -10000.0  # <<< TODO: set this to the distance to sea surface <<<
        flipFlag = True
    else:
        flipFlag = False

    # reverse line in the solver if end A is above end B
    if ZF < 0:
        ZF = -ZF
        reverseFlag = True
    else:
        reverseFlag = False

    # ensure the input variables are realistic
    if XF < 0.0:
        raise CatenaryError("XF is negative!")
    if L <= 0.0:
        breakpoint()
        raise CatenaryError("L is zero or negative!")
    if EA <= 0.0:
        raise CatenaryError("EA is zero or negative!")

    # Solve for the horizontal and vertical forces at the fairlead (HF, VF) and at the anchor (HA, VA)

    # There are many "ProfileTypes" of a mooring line and each must be analyzed separately (1-3 are consistent with FAST v7)
    # ProfileType=0: Entire line is on seabed
    # ProfileType=1: No portion of the line rests on the seabed
    # ProfileType=2: A portion of the line rests on the seabed and the anchor tension is nonzero
    # ProfileType=3: A portion of the line must rest on the seabed and the anchor tension is zero
    # ProfileType=4: The line is negatively buoyant, seabed interaction is enabled, and the line
    # is longer than a full L between end points (including stretching) i.e. it is horizontal
    # along the seabed from the anchor, then vertical to the fairlead. Computes the maximum
    # stretched length of the line with seabed interaction beyond which the line would have to
    # double-back on itself; the line forms an "L" between the anchor and fairlead. Then it
    # models it as bunched up on the seabed (instead of throwing an error)
    # ProfileType=5: Similar to above but both ends are off seabed, so it's U shaped and fully slack
    # ProfileType=6: Completely vertical line that is off the seabed (on the seabed is handled by 4 and 5)

    EA_W = EA / W

    # calculate the unstretched length that would be hanging if the line was fully slack (vertical down to flat on the seabed)
    if CB < 0:  # free floating (potentially U shaped case)
        LHanging1 = np.sqrt(2.0 * (-CB) * EA_W + EA_W * EA_W) - EA_W  # unstretched hanging length at end A
        LHanging2 = np.sqrt(2.0 * (ZF - CB) * EA_W + EA_W * EA_W) - EA_W  # unstretched hanging length at end B
        LHanging = LHanging1 + LHanging2
    else:  # at least one end on seabed
        LHanging = (
            np.sqrt(2.0 * ZF * EA_W + EA_W * EA_W) - EA_W
        )  # unstretched length of line hanging vertically to seabed

    # calculate a vertical stiffness estimate for an end lifting off the seabed
    def dV_dZ_s(z0, H):  # height off seabed to evaluate at (infinite if 0), horizontal tension
        # return W*(z0*W/H + 1)/np.sqrt( (z0*W/H + 1)**2 - 1)   # inelastic apprxoimation
        return W  # returning a fully slack line approximation,
        #   because a large value here risks adding a bad cross coupling term in the system stiffness matrix

    # ProfileType 0 case - entirely along seabed
    if ZF == 0.0 and CB >= 0.0 and W > 0:

        ProfileType = 0

        if CB == 0 or XF <= L:  # case 1: no friction, or zero tension
            HF = np.max([0, (XF / L - 1.0) * EA])
            HA = 1.0 * HF
        elif (
            0.5 * L + EA / CB / W * (1 - XF / L) <= 0
        ):  # case 2: seabed friction but tension at anchor (xB estimate < 0)
            HF = (XF / L - 1.0) * EA + 0.5 * CB * W * L
            HA = np.max([0.0, HF - CB * W * L])
        else:  # case 3: seabed friction and zero anchor tension
            HF = np.sqrt(2 * EA * CB * W * (XF - L))
            HA = 0.0

        VF = 0.0
        VA = 0.0

        if HF > 0:  # if taut
            dHF_dXF = EA / L  # approximation <<<  what about friction?  <<<<<<<<
            # dVF_dZF = W + HF/L # vertical stiffness <<< approximation a
            dVF_dZF = dV_dZ_s(Tol, HF)  # vertical stiffness <<< approximation b
        else:  # if slack
            dHF_dXF = 0.0
            dVF_dZF = W  # vertical stiffness

        info[
            "HF"
        ] = HF  # solution to be used to start next call (these are the solved variables, may be for anchor if line is reversed)
        info["VF"] = 0.0
        info["stiffnessB"] = np.array([[dHF_dXF, 0.0], [0.0, dVF_dZF]])
        info["stiffnessA"] = np.array([[dHF_dXF, 0.0], [0.0, dVF_dZF]])
        info["stiffnessAB"] = np.array([[-dHF_dXF, 0.0], [0.0, 0.0]])
        info["LBot"] = L
        info["ProfileType"] = 0
        info["Zextreme"] = 0

        if plots > 0:

            if CB > 0 and XF > L:
                xB = L - HF / W / CB  # location of point at which line tension reaches zero
            else:
                xB = 0.0

            # z values remain zero in this case

            if CB == 0 or XF <= L:  # case 1: no friction, or zero tension
                Xs = XF / L * s  # X values uniformly distributed
                Te = Te + np.max([0, (XF / L - 1.0) * EA])  # uniform tension
            elif xB <= 0:  # case 2: seabed friction but tension at anchor
                Xs = s * (1 + CB * W / EA * (0.5 * s - xB))
                Te = HF + CB * W * (s - L)
            else:  # case 3: seabed friction and zero anchor tension
                for I in range(nNodes):
                    if s[I] <= xB:  # if this node is in the zero tension range
                        Xs[I] = s[I]
                        # x is unstretched, z and Te remain zero

                    else:  # the tension is nonzero
                        Xs[I] = s[I] + CB * W / EA * (s[I] - xB) ** 2
                        Te[I] = HF - CB * W * (L - s[I])

    # ProfileType 4 case - fully slack
    elif (W > 0.0) and (L >= XF + LHanging):

        if CB >= 0.0:  # one end on seabed
            ProfileType = 4
            # this is a special case that requires no iteration

            HF = 0.0
            VF = W * LHanging
            HA = 0.0
            VA = 0.0

            dVF_dZF = W / np.sqrt(2.0 * ZF / EA_W + 1.0)  # vertical stiffness

            info[
                "HF"
            ] = HF  # solution to be used to start next call (these are the solved variables, may be for anchor if line is reversed)
            info["VF"] = VF
            info["stiffnessB"] = np.array([[0.0, 0.0], [0.0, dVF_dZF]])
            info["stiffnessA"] = np.array([[0.0, 0.0], [0.0, W]])
            info["stiffnessAB"] = np.array([[0.0, 0.0], [0.0, 0.0]])
            info["LBot"] = L - LHanging
            info["ProfileType"] = 4
            info["Zextreme"] = 0

            if plots > 0:

                for I in range(nNodes):
                    if s[I] > L - LHanging:  # this node is on the suspended/hanging portion of the line

                        Xs[I] = XF
                        Zs[I] = ZF - (L - s[I] + 0.5 * W / EA * (L - s[I]) ** 2)  # <<<< double check this
                        Te[I] = W * (L - s[I])

                    else:  # this node is on the seabed

                        Xs[I] = np.min([s[I], XF])
                        Zs[I] = 0.0
                        Te[I] = 0.0

        else:  # U shaped
            ProfileType = 5

            HF = 0.0
            VF = W * LHanging2
            HA = 0.0
            VA = -W * LHanging1

            dVF_dZF = W / np.sqrt(2.0 * ZF / EA_W + 1.0)  # vertical stiffness

            info[
                "HF"
            ] = HF  # solution to be used to start next call (these are the solved variables, may be for anchor if line is reversed)
            info["VF"] = VF
            info["stiffnessB"] = np.array([[0.0, 0.0], [0.0, W / np.sqrt(2.0 * (ZF - CB) / EA_W + 1.0)]])
            info["stiffnessA"] = np.array([[0.0, 0.0], [0.0, W / np.sqrt(2.0 * (-CB) / EA_W + 1.0)]])
            info["stiffnessAB"] = np.array([[0.0, 0.0], [0.0, 0.0]])
            info["LBot"] = L - LHanging
            info["ProfileType"] = 5
            info["Zextreme"] = CB

            if plots > 0:

                for I in range(nNodes):
                    if s[I] < LHanging1:  # the 1st suspended/hanging portion of the line
                        Xs[I] = 0.0
                        Zs[I] = -s[I] - W / EA * (LHanging1 * s[I] - 0.5 * s[I] ** 2)
                        Te[I] = W * s[I]

                    elif s[I] <= L - LHanging2:  # the middle portion of the line, slack along the seabed
                        Xs[I] = (s[I] - LHanging1) * XF / (L - LHanging1 - LHanging2)
                        Zs[I] = CB
                        Te[I] = 0.0

                    else:  # the 2nd suspended/hanging portion of the line
                        Lms = L - s[I]  # distance from end B
                        Xs[I] = XF
                        Zs[I] = ZF - Lms - W / EA * (LHanging2 * Lms - 0.5 * Lms**2)
                        Te[I] = W * Lms

    # ProfileType 6 case - vertical line without seabed contact
    elif XF == 0:
        ProfileType = 6

        dz_hanging = L + 0.5 * W / EA * L**2  # stretched length if it was hanging from one end

        # slack case
        if dz_hanging >= ZF:

            # figure out how line will hang
            LB = (ZF + L + W * L**2 / 2 / EA) / (
                2 + W * L / EA
            )  # unstretched length of line from lowest point up to end B
            hB = LB + W / 2 / EA * LB**2  # stretched of the above
            LA = L - LB
            hA = hB - ZF

            HF = 0.0
            VF = W * LB
            HA = 0.0
            VA = W * LA

            info[
                "HF"
            ] = HF  # solution to be used to start next call (these are the solved variables, may be for anchor if line is reversed)
            info["VF"] = VF
            info["stiffnessB"] = np.array([[0.0, 0.0], [0.0, 0.5 * W]])
            info["stiffnessA"] = np.array([[0.0, 0.0], [0.0, 0.5 * W]])
            info["stiffnessAB"] = np.array([[-0.0, 0.0], [0.0, -0.5 * W]])
            info["LBot"] = 0.0
            info["ProfileType"] = 6
            info["Zextreme"] = -hA

            if plots > 0:

                for I in range(nNodes):
                    if s[I] < LA:  # the 1st suspended/hanging portion of the line
                        Xs[I] = 0.0
                        Zs[I] = -s[I] - W / EA * (LA * s[I] - 0.5 * s[I] ** 2)
                        Te[I] = W * (LA - s[I])
                    else:  # the 2nd suspended/hanging portion of the line
                        Lms = L - s[I]  # distance from end B
                        Xs[I] = XF
                        Zs[I] = ZF - Lms - W / EA * (LB * Lms - 0.5 * Lms**2)
                        Te[I] = W * (s[I] - LA)

        # taut case
        else:

            # figure out how line will hang
            LB = (ZF + L + W * L**2 / 2 / EA) / (
                2 + W * L / EA
            )  # unstretched length of line from lowest point up to end B
            hB = LB + W / 2 / EA * LB**2  # stretched of the above
            LA = L - LB
            hA = hB - ZF

            uniform_strain = (
                ZF - dz_hanging
            ) / L  # the constrant strain due only to stretch - to be added to weight-based strain
            Tstretch = uniform_strain * EA  # the constant tension component to be added to weight-based tension

            HF = 0.0
            VF = Tstretch + W * L
            HA = 0.0
            VA = Tstretch

            info[
                "HF"
            ] = HF  # solution to be used to start next call (these are the solved variables, may be for anchor if line is reversed)
            info["VF"] = VF
            info["stiffnessB"] = np.array([[Tstretch / ZF, 0.0], [0.0, EA / L]])
            info["stiffnessA"] = np.array([[Tstretch / ZF, 0.0], [0.0, EA / L]])
            info["stiffnessAB"] = np.array([[-Tstretch / ZF, 0.0], [0.0, -EA / L]])
            info["LBot"] = 0.0
            info["ProfileType"] = 6
            info["Zextreme"] = 0

            if plots > 0:
                for I in range(nNodes):
                    Lms = L - s[I]  # distance from end B
                    Xs[I] = 0.0
                    Zs[I] = ZF - Lms * L * uniform_strain - W / EA * (L * Lms - 0.5 * Lms**2)
                    Te[I] = Tstretch + W * s[I]

    # Use an iterable solver function to solve for the forces on the line
    else:

        # Initialize some commonly used terms that don't depend on the iteration:

        WL = W * L
        WEA = W * EA
        L_EA = L / EA
        CB_EA = CB / EA
        # MaxIter = 50 #int(1.0/Tol)   # Smaller tolerances may take more iterations, so choose a maximum inversely proportional to the tolerance

        # more initialization
        I = 1  # Initialize iteration counter
        FirstIter = 1  # 1 means first attempt (can be retried), 0 means it's alread been retried, -1 triggers a retry

        # make HF and VF initial guesses if either was provided as zero <<<<<<<<<<<< why does it matter if VF0 is zero??
        if HF0 <= 0 or VF0 <= 0:

            XF2 = XF * XF
            ZF2 = ZF * ZF

            if L <= np.sqrt(XF2 + ZF2):  # if the current mooring line is taut
                Lamda0 = 0.2
            else:  # The current mooring line must be slack and not vertical
                Lamda0 = np.sqrt(3.0 * ((L * L - ZF2) / XF2 - 1.0))

            HF = np.max([abs(0.5 * W * XF / Lamda0), Tol])
            # ! As above, set the lower limit of the guess value of HF to the tolerance
            VF = 0.5 * W * (ZF / np.tanh(Lamda0) + L)
        else:
            HF = 1.0 * HF0
            VF = 1.0 * VF0

        # >>> note, the above Tol uses should be adjusted now that I've changed it to be absolute and distance <<<

        # make sure required values are non-zero
        HF = np.max([HF, Tol])
        XF = np.max([XF, Tol])
        ZF = np.max([ZF, Tol])

        # some initial values just for printing before they're filled in
        EXF = 0
        EZF = 0

        # Solve the analytical, static equilibrium equations for a catenary (or taut) mooring line with seabed interaction:
        X0 = [HF, VF]
        Ytarget = [0, 0]
        args = dict(cat=[XF, ZF, L, EA, W, CB, WL, WEA, L_EA, CB_EA], step=[0.15, 1.0, 1.5])
        # call the master solver function
        # X, Y, info2 = msolve.dsolve(eval_func_cat, X0, Ytarget=Ytarget, step_func=step_func_cat, args=args, tol=Tol, maxIter=MaxIter, a_max=1.2)
        X, Y, info2 = dsolve2(
            eval_func_cat,
            X0,
            Ytarget=Ytarget,
            step_func=step_func_cat,
            args=args,
            ytol=Tol,
            stepfac=1,
            maxIter=MaxIter,
            a_max=1.2,
        )

        # retry if it failed
        if info2["iter"] >= MaxIter - 1 or info2["oths"]["error"] == True or np.linalg.norm(info2["err"]) > 10 * Tol:
            #  ! Perhaps we failed to converge because our initial guess was too far off.
            #   (This could happen, for example, while linearizing a model via large
            #   pertubations in the DOFs.)  Instead, use starting values documented in:
            #   Peyrot, Alain H. and Goulois, A. M., "Analysis Of Cable Structures,"
            #   Computers & Structures, Vol. 10, 1979, pp. 805-813:
            # NOTE: We don't need to check if the current mooring line is exactly
            #       vertical (i.e., we don't need to check if XF == 0.0), because XF is
            #       limited by the tolerance above. */

            if (
                info2["iter"] >= MaxIter - 1 and XF / ZF < 0.001
            ):  # if it's nearly vertical, keep iterating from the last point
                HF = X[0]
                VF = X[1]
            else:  # otherwise try starting from some good initial guesses
                if L <= np.sqrt(XF**2 + ZF**2):  # if the current mooring line is taut
                    Lamda0 = 0.2
                else:  # The current mooring line must be slack and not vertical
                    Lamda0 = np.sqrt(3.0 * ((L * L - ZF**2) / XF**2 - 1.0))

                HF = np.max(
                    [abs(0.5 * W * XF / Lamda0), Tol]
                )  # As above, set the lower limit of the guess value of HF to the tolerance
                VF = 0.5 * W * (ZF / np.tanh(Lamda0) + L)

            X0 = [HF, VF]
            Ytarget = [0, 0]
            args = dict(
                cat=[XF, ZF, L, EA, W, CB, WL, WEA, L_EA, CB_EA], step=[0.1, 0.8, 1.5]
            )  # step: alpha_min, alpha0, alphaR
            # call the master solver function
            # X, Y, info3 = msolve.dsolve(eval_func_cat, X0, Ytarget=Ytarget, step_func=step_func_cat, args=args, tol=Tol, maxIter=MaxIter, a_max=1.1) #, dX_last=info2['dX'])
            X, Y, info3 = dsolve2(
                eval_func_cat,
                X0,
                Ytarget=Ytarget,
                step_func=step_func_cat,
                args=args,
                ytol=Tol,
                stepfac=1,
                maxIter=MaxIter,
                a_max=1.2,
            )

            # retry if it failed
            if info3["iter"] >= MaxIter - 1 or info3["oths"]["error"] == True:

                X0 = X
                Ytarget = [0, 0]
                args = dict(cat=[XF, ZF, L, EA, W, CB, WL, WEA, L_EA, CB_EA], step=[0.1, 1.0, 2.0])
                # call the master solver function
                # X, Y, info4 = msolve.dsolve(eval_func_cat, X0, Ytarget=Ytarget, step_func=step_func_cat, args=args, tol=Tol, maxIter=10*MaxIter, a_max=1.15) #, dX_last=info3['dX'])
                X, Y, info4 = dsolve2(
                    eval_func_cat,
                    X0,
                    Ytarget=Ytarget,
                    step_func=step_func_cat,
                    args=args,
                    ytol=Tol,
                    stepfac=1,
                    maxIter=MaxIter,
                    a_max=1.2,
                )

                # check if it failed
                if info4["iter"] >= 10 * MaxIter - 1 or info4["oths"]["error"] == True:

                    print("catenary solve failed on all 3 attempts.")
                    print(
                        f"catenary({XF}, {ZF}, {L}, {EA}, {W}, CB={CB}, HF0={HF0}, VF0={VF0}, Tol={Tol}, MaxIter={MaxIter}, plots=1)"
                    )

                    print("First attempt's iterations are as follows:")
                    for i in range(info2["iter"] + 1):
                        print(
                            f" Iteration {i}: HF={info2['Xs'][i,0]: 8.4e}, VF={info2['Xs'][i,1]: 8.4e}, EX={info2['Es'][i,0]: 6.2e}, EZ={info2['Es'][i,1]: 6.2e}"
                        )

                    print("Second attempt's iterations are as follows:")
                    for i in range(info3["iter"] + 1):
                        print(
                            f" Iteration {i}: HF={info3['Xs'][i,0]: 8.4e}, VF={info3['Xs'][i,1]: 8.4e}, EX={info3['Es'][i,0]: 6.2e}, EZ={info3['Es'][i,1]: 6.2e}"
                        )

                    print("Last attempt's iterations are as follows:")
                    for i in range(info4["iter"] + 1):
                        print(
                            f" Iteration {i}: HF={info4['Xs'][i,0]: 8.4e}, VF={info4['Xs'][i,1]: 8.4e}, EX={info4['Es'][i,0]: 6.2e}, EZ={info4['Es'][i,1]: 6.2e}"
                        )

                    """
                    # plot solve performance
                    fig, ax = plt.subplots(4,1, sharex=True)
                    ax[0].plot(np.hstack([info2['Xs'][:,0], info3['Xs'][:,0], info4['Xs'][:,0]]))
                    ax[1].plot(np.hstack([info2['Xs'][:,1], info3['Xs'][:,1], info4['Xs'][:,1]]))
                    ax[2].plot(np.hstack([info2['Es'][:,0], info3['Es'][:,0], info4['Es'][:,0]]))
                    ax[3].plot(np.hstack([info2['Es'][:,1], info3['Es'][:,1], info4['Es'][:,1]]))
                    ax[0].set_ylabel("HF")
                    ax[1].set_ylabel("VF")
                    ax[2].set_ylabel("X err")
                    ax[3].set_ylabel("Z err")

                    # plot solve path
                    plt.figure()

                    #c = np.hypot(info2['Es'][:,0], info2['Es'][:,1])


                    c = np.arange(info2['iter']+1)
                    c = cm.jet((c-np.min(c))/(np.max(c)-np.min(c)))

                    for i in np.arange(info2['iter']):
                        plt.plot(info2['Xs'][i:i+2,0], info2['Xs'][i:i+2,1],":", c=c[i])
                    plt.plot(info2['Xs'][0,0], info2['Xs'][0,1],"o")

                    c = np.arange(info3['iter']+1)
                    c = cm.jet((c-np.min(c))/(np.max(c)-np.min(c)))

                    for i in np.arange(info3['iter']):
                        plt.plot(info3['Xs'][i:i+2,0], info3['Xs'][i:i+2,1], c=c[i])
                    plt.plot(info3['Xs'][0,0], info3['Xs'][0,1],"*")

                    c = np.arange(info4['iter']+1)
                    c = cm.jet((c-np.min(c))/(np.max(c)-np.min(c)))

                    for i in np.arange(info4['iter']):
                        plt.plot(info4['Xs'][i:i+2,0], info4['Xs'][i:i+2,1], c=c[i])
                    plt.plot(info4['Xs'][0,0], info4['Xs'][0,1],"*")

                    plt.title("catenary solve path for troubleshooting")
                    plt.show()

                    #breakpoint()
                    """
                    raise CatenaryError("catenary solver failed.")

                else:  # if the solve was successful,
                    info.update(info4["oths"])  # copy info from last solve into existing info dictionary
                    info["catenary"] = info4

            else:  # if the solve was successful,
                info.update(info3["oths"])  # copy info from last solve into existing info dictionary
                info["catenary"] = info3
        else:  # if the solve was successful,
            info.update(info2["oths"])  # copy info from last solve into existing info dictionary
            info["catenary"] = info2

        # check for errors ( WOULD SOME NOT ALREADY HAVE BEEN CAUGHT AND RAISED ALREADY?)
        if info["error"] == True:
            # breakpoint()
            # >>>> what about errors for which we can first plot the line profile?? <<<<
            raise CatenaryError("Error in catenary computations: " + info["message"])

        # if info['Zextreme'] < CB:
        #    info["warning"] = "Line is suspended from both ends but hits the seabed (this isn't allowed in MoorPy)"

        ProfileType = info["ProfileType"]
        HF = X[0]
        VF = X[1]
        HA = info["HA"]
        VA = info["VA"]

        # --- now that the iterative solve is over, check some things on the results, handle plotting, etc. ---

        # compute the Zextreme value - for a freely suspended line, if necessary, check to ensure the line doesn't droop and hit the seabed
        if (
            info["ProfileType"] == 1 and CB < 0 and VF - WL < 0.0
        ):  # only need to do this if the line is slack (has zero slope somewhere)

            VFMinWL = VF - WL
            LBot = L - VF / W
            # unstretched length of line resting on seabed (Jonkman's PhD eqn 2-38), LMinVFOVrW
            HF_W = HF / W
            HF_WEA = HF / WEA
            VF_WEA = VF / WEA
            VF_HF = VF / HF
            VFMinWL_HF = VFMinWL / HF
            VF_HF2 = VF_HF * VF_HF
            VFMinWL_HF2 = VFMinWL_HF * VFMinWL_HF
            SQRT1VF_HF2 = np.sqrt(1.0 + VF_HF2)
            SQRT1VFMinWL_HF2 = np.sqrt(1.0 + VFMinWL_HF2)

            # this is indicated by the anchor force having a positive value, meaning it's helping hold up the line
            info["Sextreme"] = L - VF / W  # arc length where slope is zero
            info["Zextreme"] = (
                1 - SQRT1VFMinWL_HF2
            ) * HF_W - 0.5 * VFMinWL**2 / WEA  # max or min line elevation (where slope=0)
            info["Xextreme"] = (-np.log(VFMinWL_HF + SQRT1VFMinWL_HF2)) * HF_W + HF * info["Sextreme"] / EA
        else:
            info["Sextreme"] = 0.0
            info["Zextreme"] = 0.0
            info["Xextreme"] = 0.0

        # handle special case of a U-shaped line that has seabed contact (using 2 new catenary solves)
        if info["ProfileType"] == 1 and info["Zextreme"] < min(CB, 0):

            # we will solve this as two separate lines to form the U shape
            info["ProfileType"] = "U"
            ProfileType = "U"

            X1_0 = info["Xextreme"]  # define fake anchor point as lowest point of line (if the seabed wasn't there)
            X2_0 = XF - X1_0
            L1 = info["Sextreme"]
            L2 = L - L1
            Z1 = CB  # negative of height from seabed to original 'anchor' end [m]
            Z2 = -Z1 + ZF  # height from seabed to fairlead end

            # set up a 1D solve for the correct choice of the anchor point so that horizontal tensions balance

            def eval_func_U(X, args):

                info = dict(error=False)

                X1 = X[0]
                X2 = XF - X1

                # note: reducing tolerances for these sub-calls <<< how much is good? <<<
                (fAH1, fAV1, fBH1, fBV1, info1) = catenary(X1, Z1, L1, EA, W, CB=0, Tol=0.5 * Tol, MaxIter=MaxIter)
                (fAH2, fAV2, fBH2, fBV2, info2) = catenary(X2, Z2, L2, EA, W, CB=0, Tol=0.5 * Tol, MaxIter=MaxIter)

                Himbalance = fBH2 - fBH1

                K1 = info1[
                    "stiffnessA"
                ]  # note: this refers to the upper end of this half of the line (since it is called with Z<0)
                K2 = info2["stiffnessB"]

                info["dH_dX"] = (
                    K1[0, 0] + K2[0, 0]
                )  # horizontal stiffness on connection point on seabed between two line portions

                # print(f" X1 = {X1}, H1 = {fBH1}, H2 = {fBH2}, err={Himbalance}, dH/dX = {info['dH_dX']}")\
                # breakpoint()

                return np.array([Himbalance]), info, False  # returns Y value, misc dict, and stop flag

            def step_func_U(X, args, Y, info, Ytarget, err, tols, iter, maxIter):

                dX = -err[0] / info["dH_dX"]

                # print(f" Step is {dX}")

                return np.array([dX])  # returns dX (step to make)

            # call this to solve for line shapes that balance the horizontal tension in the line
            X, Y, infoU = dsolve2(
                eval_func_U, [X1_0], step_func=step_func_U, ytol=0.25 * Tol, stepfac=1, maxIter=20, a_max=1.2, display=0
            )
            X1 = X[0]
            X2 = XF - X1

            # call one more time to get final values
            (fAH1, fAV1, fBH1, fBV1, info1) = catenary(
                X1, Z1, L1, EA, W, CB=0, Tol=0.5 * Tol, MaxIter=MaxIter, plots=plots
            )
            (fAH2, fAV2, fBH2, fBV2, info2) = catenary(
                X2, Z2, L2, EA, W, CB=0, Tol=0.5 * Tol, MaxIter=MaxIter, plots=plots
            )

            if plots > 0 or (info1["error"] and info2["error"]):

                s = np.hstack([info1["s"], info2["s"] + L1])
                Xs = np.hstack([info1["X"], info2["X"] + X1])
                Zs = np.hstack([info1["Z"], info2["Z"] + Z1])
                Te = np.hstack([info1["Te"], info2["Te"]])

                # re-reverse line distributed data back to normal if applicable
                """
                if reverseFlag:
                    info['s']  =  L - info['s' ][::-1]
                    info['X']  = XF - info['X' ][::-1]
                    info['Z']  =      info['Z' ][::-1] - ZF  # remember ZF still has a flipped sign right now
                    info['Te'] =      info['Te'][::-1]
                """

            if flipFlag:
                raise Exception(
                    "flipFlag connot be True for the case of a U shaped line with seabed contact. Something must be wrong."
                )

            # get stiffnesses    (check sign of A!)
            K1 = info1[
                "stiffnessA"
            ]  # note: this refers to the upper end of this half of the line (since it is called with Z<0)
            K2 = info2["stiffnessB"]
            dH_dX = 1.0 / (1.0 / K1[0, 0] + 1.0 / K2[0, 0])  # = K1[0,0]*K2[0,0]/(K1[0,0] + K2[0,0])
            Kmid = K1[0, 0] + K2[0, 0]  # horizontal stiffness on connection point on seabed between two line portions

            dxdH = 1.0 / Kmid  # = 1/(K1[0,0] + K2[0,0])

            info["stiffnessA"] = np.array(
                [
                    [dH_dX, K1[0, 1] * K2[0, 0] * dxdH],
                    [K1[1, 0] * K2[0, 0] * dxdH, K1[1, 1] - K1[1, 0] * dxdH * K1[0, 1]],
                ]
            )

            info["stiffnessB"] = np.array(
                [
                    [dH_dX, K2[0, 1] * K1[0, 0] * dxdH],
                    [K2[1, 0] * K1[0, 0] * dxdH, K2[1, 1] - K2[1, 0] * dxdH * K2[0, 1]],
                ]
            )

            info["stiffnessAB"] = np.array(
                [
                    [
                        -K1[0, 0] * K2[0, 0] * dxdH,
                        -K1[0, 1] * K2[0, 0] * dxdH,
                    ],  # this is the lower-left submatrix, A motions, B reaction forces
                    [-K1[0, 0] * dxdH * K2[1, 0], -K1[0, 1] * dxdH * K2[1, 0]],
                ]
            )

            #                         xA                              zA                             xB                                 zB

            info["K"] = np.array(
                [
                    [
                        K1[0, 0] * K2[0, 0] * dxdH,
                        K1[0, 1] * K2[0, 0] * dxdH,
                        -K2[0, 0] * K1[0, 0] * dxdH,
                        -K2[0, 1] * K1[0, 0] * dxdH,
                    ],  # HA
                    [
                        K1[1, 0] * K2[0, 0] * dxdH,
                        K1[1, 1] - K1[1, 0] * dxdH * K1[0, 1],
                        -K2[0, 0] * dxdH * K1[1, 0],
                        -K2[0, 1] * dxdH * K1[1, 0],
                    ],  # VA
                    [
                        -K1[0, 0] * K2[0, 0] * dxdH,
                        -K1[0, 1] * K2[0, 0] * dxdH,
                        K2[0, 0] * K1[0, 0] * dxdH,
                        K2[0, 1] * K1[0, 0] * dxdH,
                    ],  # HB
                    [
                        -K1[0, 0] * dxdH * K2[1, 0],
                        -K1[0, 1] * dxdH * K2[1, 0],
                        K2[1, 0] * K1[0, 0] * dxdH,
                        K2[1, 1] - K2[1, 0] * dxdH * K2[0, 1],
                    ],
                ]
            )  # VB

            """

                            \frac{  \pderiv{H_A}{x_A}\pderiv{H_B}{x_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}}
                            &\frac{ \pderiv{H_A}{z_A}\pderiv{H_B}{x_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}}
                            &-\frac{\pderiv{H_A}{x_A}\pderiv{H_B}{x_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}}
                            &-\frac{\pderiv{H_A}{x_A}\pderiv{H_B}{z_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}} \\

                            \frac{  \pderiv{V_A}{x_A}\pderiv{H_B}{x_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}}
                            &\pderiv{V_A}{z_A} - \frac{ \pderiv{H_A}{z_A}\pderiv{V_A}{x_A}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}}
                            &-\frac{\pderiv{V_A}{x_A}\pderiv{H_B}{x_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}}
                            &-\frac{\pderiv{V_A}{x_A}\pderiv{H_B}{z_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}} \\

                            -\frac{ \pderiv{H_A}{x_A}\pderiv{H_B}{x_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}}
                            &-\frac{\pderiv{H_A}{z_A}\pderiv{H_B}{x_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}}
                            &\frac{ \pderiv{H_A}{x_A}\pderiv{H_B}{x_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}}
                            &\frac{ \pderiv{H_A}{x_A}\pderiv{H_B}{z_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}} \\

                            -\frac{ \pderiv{H_A}{x_A}\pderiv{V_B}{x_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}}
                            &-\frac{\pderiv{H_A}{z_A}\pderiv{V_B}{x_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}}
                            &\frac{ \pderiv{H_A}{x_A}\pderiv{V_B}{x_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}}
                            &\pderiv{V_B}{z_B} - \frac{ \pderiv{H_B}{z_B}\pderiv{V_B}{x_B}}{\pderiv{H_A}{x_A}+\pderiv{H_B}{x_B}}



                            for a normal line

                            \pderiv{H_B}{x_B}   &?                  & -\pderiv{H_B}{x_B} & \pderiv{H_B}{z_B}\\  # HA
                            ?                   & ?                 &      0             &   0              \\  # VA
                            -\pderiv{H_B}{x_B}+  &    0ish          & \pderiv{H_B}{x_B}  & \pderiv{H_B}{z_B}\\  # HB
                            -\pderiv{V_B}{x_B}+  &    0ish          & \pderiv{V_B}{x_B}  & \pderiv{V_B}{z_B}    # VB


            # sorted
            K  =  np.array([[                  dH_dX,    K1[0,1] *K2[0,0]*dxdH        ,                   -dH_dX,   -K2[0,1] *K1[0,0]*dxdH          ],  # HA
                            [  K1[1,0] *K2[0,0]*dxdH,    K1[1,1] -K1[1,0]*dxdH*K1[0,1],   -K2[0,0] *dxdH*K1[1,0],   -K2[0,1] *dxdH*K1[1,0]          ],  # VA
                            [                 -dH_dX,   -K1[0,1] *K2[0,0]*dxdH        ,                    dH_dX,    K2[0,1] *K1[0,0]*dxdH          ],  # HB
                            [ -K1[0,0] *dxdH*K2[1,0],   -K1[0,1] *dxdH*K2[1,0]        ,    K2[1,0] *K1[0,0]*dxdH,    K2[1,1] -K2[1,0]*dxdH*K2[0,1]  ]]) # VB
            """

            info["LBot"] = info1["LBot"] + info2["LBot"]
            # not very useful outputs for this case:
            info["Sextreme"] = L1 - info1["LBot"]
            info["Zextreme"] = CB
            info["Xextreme"] = X1 - info1["LBot"]

            # FxA = fAH1
            # FzA = fAV1
            # FxB = fBH2
            # FzB = fBV2
            HA = fAH1
            VA = fAV1
            HF = -fBH2
            VF = -fBV2

            if plots > 3:
                plt.plot(Xs, Zs)
                plt.show()

        # the normal case
        else:

            # do plotting-related calculations if needed (plots=1: show plots; plots=2: just return values)
            if plots > 0 or info["error"] == True:

                # calculate some commonly used terms that depend on HF and VF:  AGAIN
                VFMinWL = VF - WL
                LBot = L - VF / W
                # unstretched length of line resting on seabed (Jonkman's PhD eqn 2-38), LMinVFOVrW
                HF_W = HF / W
                # HF_WEA           =      HF/WEA
                # VF_WEA           =      VF/WEA
                VF_HF = VF / HF
                VFMinWL_HF = VFMinWL / HF
                VF_HF2 = VF_HF * VF_HF
                # VFMinWL_HF2      = VFMinWL_HF*VFMinWL_HF
                # SQRT1VF_HF2      = np.sqrt( 1.0 + VF_HF2      )
                SQRT1VFMinWL_HF2 = np.sqrt(1.0 + VFMinWL_HF**2)

                for I in range(nNodes):

                    # calculate some values for the current node
                    Ws = W * s[I]
                    VFMinWLs = VFMinWL + Ws  # = VF - W*(L-s[I])
                    VFMinWLs_HF = VFMinWLs / HF
                    s_EA = s[I] / EA
                    SQRT1VFMinWLs_HF2 = np.sqrt(1.0 + VFMinWLs_HF * VFMinWLs_HF)

                    # No portion of the line rests on the seabed
                    if ProfileType == 1:

                        Xs[I] = (
                            np.log(VFMinWLs_HF + SQRT1VFMinWLs_HF2) - np.log(VFMinWL_HF + SQRT1VFMinWL_HF2)
                        ) * HF_W + s_EA * HF
                        Zs[I] = (SQRT1VFMinWLs_HF2 - SQRT1VFMinWL_HF2) * HF_W + s_EA * (VFMinWL + 0.5 * Ws)
                        Te[I] = np.sqrt(HF * HF + VFMinWLs * VFMinWLs)

                    # A portion of the line must rest on the seabed
                    elif ProfileType in [2, 3]:

                        if CB > 0:
                            xB = LBot - HF_W / CB  # location of point at which line tension reaches zero
                        else:
                            xB = 0.0
                        xBlim = max(xB, 0.0)

                        if (
                            s[I] <= xB and CB > 0
                        ):  # (aka Lbot - s > HF/(CB*W) ) if this node rests on the seabed and the tension is zero

                            Xs[I] = s[I]
                            Zs[I] = 0.0
                            Te[I] = 0.0

                        elif s[I] <= LBot:  # // .TRUE. if this node rests on the seabed and the tension is nonzero

                            Xs[I] = s[I] + 0.5 * CB * W / EA * (s[I] * s[I] - 2.0 * xB * s[I] + xB * xBlim)
                            Zs[I] = 0.0
                            Te[I] = HF + CB * VFMinWLs

                        else:  #  // LBot < s <= L ! This node must be above the seabed

                            Xs[I] = (
                                LBot
                                + HF_W * np.log(VFMinWLs_HF + SQRT1VFMinWLs_HF2)
                                + HF * s_EA
                                + 0.5 * CB * W / EA * (-LBot * LBot + xB * xBlim)
                            )
                            Zs[I] = (
                                (-1.0 + SQRT1VFMinWLs_HF2) * HF_W
                                + s_EA * (VFMinWL + 0.5 * Ws)
                                + 0.5 * VFMinWL * VFMinWL / WEA
                            )
                            Te[I] = np.sqrt(HF * HF + VFMinWLs * VFMinWLs)

    if plots > 0:
        # re-reverse line distributed data back to normal if applicable
        if reverseFlag:
            s = L - s[::-1]
            Xs = XF - Xs[::-1]
            Zs = Zs[::-1] - ZF  # remember ZF still has a flipped sign right now
            Te = Te[::-1]
        if flipFlag:
            Zs = -Zs  # flip calculated line Z coordinates (hopefully this is right)

        # save data to info dict
        info["X"] = Xs
        info["Z"] = Zs
        info["s"] = s
        info["Te"] = Te

    if plots == 2 or info["error"] == True:  # also show the profile plot

        plt.figure()
        plt.plot(Xs, Zs)

    # get A and AB stiffness matrices for catenary profiles here based on fairlead (B) stiffness matrix
    if ProfileType == 1:
        info["stiffnessA"] = np.array(info["stiffnessB"])
        info["stiffnessAB"] = -info["stiffnessB"]

    elif ProfileType in [2, 3]:
        if CB == 0.0:
            info["stiffnessA"] = np.array(
                [[info["stiffnessB"][0, 0], 0], [0, dV_dZ_s(Tol, HF)]]
            )  # vertical term is very approximate
            info["stiffnessAB"] = np.array(
                [[-info["stiffnessB"][0, 0], 0], [0, 0]]
            )  # note: A and AB stiffnesses for this case only valid if zero friction
        else:
            info["stiffnessA"] = np.ones([2, 2]) * np.nan  # if friction, flag to ensure users don't use this
            info["stiffnessAB"] = np.ones([2, 2]) * np.nan  # if friction, flag to ensure users don't use this

    # un-swap line ends if they've been previously swapped, and apply global sign convention
    # (vertical force positive-up, horizontal force positive from A to B)
    if reverseFlag:
        ZF = -ZF  # put height rise from end A to B back to negative

        FxA = HF
        FzA = -VF  # VF is positive-down convention so flip sign
        FxB = -HA
        FzB = VA

        info["stiffnessA"], info["stiffnessB"] = info["stiffnessB"], info["stiffnessA"]  # swap A and B
        # note: diagonals of AB matrix do not change

        info["stiffnessA"][0, 1] = -info["stiffnessA"][0, 1]  # reverse off-diagonal signs
        info["stiffnessA"][1, 0] = -info["stiffnessA"][1, 0]
        info["stiffnessB"][0, 1] = -info["stiffnessB"][0, 1]
        info["stiffnessB"][1, 0] = -info["stiffnessB"][1, 0]
        info["stiffnessAB"][0, 1] = -info["stiffnessAB"][
            0, 1
        ]  # for cross coupling matrix could also maybe transpose? but should be symmetrical so no need
        info["stiffnessAB"][1, 0] = -info["stiffnessAB"][1, 0]

    else:
        FxA = HA
        FzA = VA
        FxB = -HF
        FzB = -VF

    if flipFlag:
        W = -W  # restore original
        ZF = -ZF  # restore original

        FzA = -FzA
        FzB = -FzB

        info["stiffnessA"], info["stiffnessB"] = info["stiffnessB"], info["stiffnessA"]  # swap A and BB
        # note: diagonals of AB matrix do not change

        info["stiffnessA"][0, 1] = -info["stiffnessA"][0, 1]  # reverse off-diagonal signs
        info["stiffnessA"][1, 0] = -info["stiffnessA"][1, 0]
        info["stiffnessB"][0, 1] = -info["stiffnessB"][0, 1]
        info["stiffnessB"][1, 0] = -info["stiffnessB"][1, 0]
        info["stiffnessAB"][0, 1] = -info["stiffnessAB"][0, 1]
        info["stiffnessAB"][1, 0] = -info["stiffnessAB"][1, 0]

        # TODO <<< should add more info <<<

    # return horizontal and vertical (positive-up) tension components at each end, and length along seabed
    return (FxA, FzA, FxB, FzB, info)


def eval_func_cat(X, args):
    """returns target outputs and also secondary outputs for constraint checks etc."""

    info = dict(error=False)  # a dict of extra outputs to be returned

    ## Step 1. break out design variables and arguments into nice names
    HF = X[0]
    VF = X[1]

    [XF, ZF, L, EA, W, CB, WL, WEA, L_EA, CB_EA] = args["cat"]

    ## Step 2. do the evaluation (this may change mutable things in args)

    # print("catenary iteration HF={:8.4e}, VF={:8.4e}".format(HF,VF))

    # calculate some commonly used terms that depend on HF and VF:

    VFMinWL = VF - WL
    # = VA, the vertical anchor load (positive-up, but VF is positive-down)
    LBot = L - VF / W
    # unstretched length of line resting on seabed (Jonkman's PhD eqn 2-38), LMinVFOVrW
    HF_W = HF / W
    HF_WEA = HF / WEA
    VF_WEA = VF / WEA
    VF_HF = VF / HF
    # VF_HF            =      np.abs(VF/HF)  # I added the abs <<<<<< <<<<<<<<<<<<<<<<<<<<<<<<<<<
    VFMinWL_HF = VFMinWL / HF
    VF_HF2 = VF_HF * VF_HF
    VFMinWL_HF2 = VFMinWL_HF * VFMinWL_HF
    SQRT1VF_HF2 = np.sqrt(1.0 + VF_HF2)
    SQRT1VFMinWL_HF2 = np.sqrt(1.0 + VFMinWL_HF2)

    # determine line profile type
    if (CB < 0.0) or (W < 0.0) or (VFMinWL > 0.0):  # no portion of the line rests on the seabed
        ProfileType = 1
    elif -CB * VFMinWL < HF:  # a portion of the line rests on the seabed and the anchor tension is nonzero
        ProfileType = 2
    else:  # must be 0.0 < HF <= -CB*VFMinWL, meaning a portion of the line must rest on the seabed and the anchor tension is zero
        ProfileType = 3

    # Compute the error functions (to be zeroed) and the Jacobian matrix
    #   (these depend on the anticipated configuration of the mooring line):

    # <<< could eliminate frequent division by W below, (make 1/W variable) >>>>>

    # No portion of the line rests on the seabed
    if ProfileType == 1:

        if VF_HF + SQRT1VF_HF2 <= 0:
            info["error"] = True
            info["message"] = "ProfileType 1: VF_HF + SQRT1VF_HF2 <= 0"
        elif VFMinWL_HF + SQRT1VFMinWL_HF2 <= 0:
            info["error"] = True
            info["message"] = "ProfileType 1: VFMinWL_HF + SQRT1VFMinWL_HF2 <= 0"
            # note: these errors seemed to occur when a buoyant line got to an HF=0 iteration (hopefully avoided now)

        else:

            LBot = 0  # note that there is no seabed contact (for clarity in outputs)

            EXF = (
                (np.log(VF_HF + SQRT1VF_HF2) - np.log(VFMinWL_HF + SQRT1VFMinWL_HF2)) * HF_W + L_EA * HF - XF
            )  # error in horizontal distance

            EZF = (SQRT1VF_HF2 - SQRT1VFMinWL_HF2) * HF_W + L_EA * (VF - 0.5 * WL) - ZF  # error in vertical distance

            dXFdHF = (
                (np.log(VF_HF + SQRT1VF_HF2) - np.log(VFMinWL_HF + SQRT1VFMinWL_HF2)) / W
                - (
                    (VF_HF + VF_HF2 / SQRT1VF_HF2) / (VF_HF + SQRT1VF_HF2)
                    - (VFMinWL_HF + VFMinWL_HF2 / SQRT1VFMinWL_HF2) / (VFMinWL_HF + SQRT1VFMinWL_HF2)
                )
                / W
                + L_EA
            )

            dXFdVF = (
                (1.0 + VF_HF / SQRT1VF_HF2) / (VF_HF + SQRT1VF_HF2)
                - (1.0 + VFMinWL_HF / SQRT1VFMinWL_HF2) / (VFMinWL_HF + SQRT1VFMinWL_HF2)
            ) / W

            dZFdHF = (SQRT1VF_HF2 - SQRT1VFMinWL_HF2) / W - (VF_HF2 / SQRT1VF_HF2 - VFMinWL_HF2 / SQRT1VFMinWL_HF2) / W

            dZFdVF = (VF_HF / SQRT1VF_HF2 - VFMinWL_HF / SQRT1VFMinWL_HF2) / W + L_EA
            # dZFdVF = ( np.sign(VF)*VF_HF /SQRT1VF_HF2 - VFMinWL_HF /SQRT1VFMinWL_HF2 )/ W + L_EA

    elif ProfileType == 27:

        if VF_HF + SQRT1VF_HF2 <= 0:
            info["error"] = True
            info["message"] = "ProfileType 2: VF_HF + SQRT1VF_HF2 <= 0"

        else:
            EXF = np.log(VF_HF + SQRT1VF_HF2) * HF_W - 0.5 * CB_EA * W * LBot * LBot + L_EA * HF + LBot - XF

            EZF = (SQRT1VF_HF2 - 1.0) * HF_W + 0.5 * VF * VF_WEA - ZF

            dXFdHF = (
                np.log(VF_HF + SQRT1VF_HF2) / W - ((VF_HF + VF_HF2 / SQRT1VF_HF2) / (VF_HF + SQRT1VF_HF2)) / W + L_EA
            )

            dXFdVF = ((1.0 + VF_HF / SQRT1VF_HF2) / (VF_HF + SQRT1VF_HF2)) / W + CB_EA * LBot - 1.0 / W

            dZFdHF = (SQRT1VF_HF2 - 1.0 - VF_HF2 / SQRT1VF_HF2) / W

            dZFdVF = (VF_HF / SQRT1VF_HF2) / W + VF_WEA
            breakpoint()

    # A portion of the line must rest on the seabed and the anchor tension is zero
    elif ProfileType in [2, 3]:

        if VF_HF + SQRT1VF_HF2 <= 0:
            info["error"] = True
            info["message"] = "ProfileType 2 or 3: VF_HF + SQRT1VF_HF2 <= 0"

        else:

            if CB > 0:
                xB = LBot - HF_W / CB  # location of point at which line tension reaches zero
            else:
                xB = 0.0
            xBlim = max(xB, 0.0)

            EXF = (
                np.log(VF_HF + SQRT1VF_HF2) * HF_W
                - 0.5 * CB_EA * W * (LBot * LBot - xBlim * xBlim)
                + L_EA * HF
                + LBot
                - XF
            )

            EZF = (SQRT1VF_HF2 - 1.0) * HF_W + 0.5 * VF * VF_WEA - ZF

            dXFdHF = (
                np.log(VF_HF + SQRT1VF_HF2) / W
                - ((VF_HF + VF_HF2 / SQRT1VF_HF2) / (VF_HF + SQRT1VF_HF2)) / W
                + L_EA
                - xBlim / EA
            )

            # dXFdVF = ( ( 1.0 + VF_HF /SQRT1VF_HF2 )/( VF_HF + SQRT1VF_HF2 ) )/ W + HF_WEA +xBlim*CB/EA- 1.0/W   <<<< incorrect, at least when CB=0
            if xB <= 0:
                dXFdVF = (
                    ((1.0 + VF_HF / SQRT1VF_HF2) / (VF_HF + SQRT1VF_HF2)) / W + CB_EA * LBot - 1.0 / W
                )  # from ProfileType 2
            else:
                dXFdVF = (
                    ((1.0 + VF_HF / SQRT1VF_HF2) / (VF_HF + SQRT1VF_HF2)) / W + HF_WEA - 1.0 / W
                )  # from ProfileType 3

            dZFdHF = (SQRT1VF_HF2 - 1.0 - VF_HF2 / SQRT1VF_HF2) / W

            dZFdVF = (VF_HF / SQRT1VF_HF2) / W + VF_WEA

    # Now compute the tensions at the anchor
    if ProfileType == 1:  # No portion of the line rests on the seabed
        HA = HF
        VA = VFMinWL  # note: VF is defined positive when tension pulls downward, while VA is defined positive when tension pulls up

    elif ProfileType == 2:  # A portion of the line rests on the seabed and the anchor tension is nonzero
        HA = (
            HF + CB * VFMinWL
        )  # note: -VFMinWL = -(VF-W*L) is the negative of line weight NOT supported by the fairlead; i.e. the weight on the seabed
        VA = 0.0

    elif ProfileType == 3:  # A portion of the line must rest on the seabed and the anchor tension is zero
        HA = 0.0
        VA = 0.0

    # if there was an error, send the stop signal
    if info["error"] == True:
        # breakpoint()
        return np.zeros(2), info, True

    ## Step 3. group the outputs into objective function value and others
    Y = np.array([EXF, EZF])  # objective function

    # info is a dict of other outputs to be returned
    info[
        "HF"
    ] = HF  # solution to be used to start next call (these are the solved variables, may be for anchor if line is reversed)
    info["VF"] = VF
    # info["jacobian"]  = np.array([[dXFdHF, dXFdVF], [dZFdHF, dZFdVF]])
    info["stiffnessB"] = np.linalg.inv(
        np.array([[dXFdHF, dXFdVF], [dZFdHF, dZFdVF]])
    )  # stiffness matrix at fairlead end
    info["LBot"] = LBot
    info["HA"] = HA
    info["VA"] = VA
    info["ProfileType"] = ProfileType

    # print("EX={:5.2e}, EZ={:5.2e}".format(EXF, EZF))

    return Y, info, False


def step_func_cat(X, args, Y, info, Ytarget, err, tols, iter, maxIter):
    """General stepping functions, which can also contain special condition checks or other adjustments to the process

    info - the info dict created by the main catenary function

    """
    [XF, ZF, L, EA, W, CB, WL, WEA, L_EA, CB_EA] = args["cat"]

    # if abs( err[1] + ZF ) < 0.0001:
    #    breakpoint()

    [alpha_min, alpha0, alphaR] = args[
        "step"
    ]  # get minimum alpha, initial alpha, and alpha reduction rate from passed arguments

    # J = info['jacobian']
    # dX = -np.matmul(np.linalg.inv(J), err)
    dX = -np.matmul(info["stiffnessB"], err)

    # ! Reduce dHF by factor (between 1 at I = 1 and 0 at I = MaxIter) that reduces linearly with iteration count
    # to ensure that we converge on a solution even in the case were we obtain a nonconvergent cycle about the
    # correct solution (this happens, for example, if we jump to quickly between a taut and slack catenary)

    alpha = 1.0  # M<<<<<<<< np.max([alpha_min, alpha0*(1.0 - alphaR*iter/maxIter)])

    # exponential approach       alpha = alpha0 * np.exp( iter/maxIter * np.log(alpha_min/alpha0 ) )

    dX[0] = dX[0] * alpha  # dHF*( 1.0 - Tol*I )
    dX[1] = dX[1] * alpha  # dVF*( 1.0 - Tol*I )

    # To avoid an ill-conditioned situation, make sure HF does not go less than or equal to zero by having a lower limit of Tol*HF
    # [NOTE: the value of dHF = ( Tol - 1.0 )*HF comes from: HF = HF + dHF = Tol*HF when dHF = ( Tol - 1.0 )*HF]
    # dX[0] = max( dX[0], ( tol - 1.0 )*info['HF']);

    # To avoid an ill-conditioned situation, make sure HF does not get too close to zero, by forcing HF >= tols[0]
    # if info['HF'] + dX[0] <= tol*abs(info['VF']+dX[1]):
    # if info['HF'] + dX[0] <= tols[0]
    if X[0] + dX[0] <= tols[0]:
        #    dX[0] = tol*abs(info['VF']+dX[1]) - info['HF']
        #    dX[0] = tols[0] - info['HF']
        dX[0] = tols[0] - X[0]

    # To avoid an ill-conditioned situation where the line is nearly all on the seabed but the solver gets stuck,
    # if np.abs(err[1] + ZF)/ZF < tol:
    #    breakpoint()
    # deltaHFVF = info['HF'] - info['VF']
    # dX[0] = dX[0] - 0.5*deltaHFVF
    # dX[1] = dX[1] + 0.5*deltaHFVF

    # prevent silly situation where a line with weight and positive ZF considers a negative VF
    if info["ProfileType"] == 2:
        if X[1] + dX[1] <= tols[1]:  # if vertical force is within tolerance of being zero/negative
            VFtarget = (L - info["LBot"]) * W  # set next VF value to be the weight of portion of line that's suspended
            dX[1] = VFtarget - X[1]

    return dX  # returns dX (step to make)


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    """
    (fAH, fAV, fBH, fBV, info) = catenary(37.96888656874307, 20.49078283711694, 100.0, 751000000.0,
                                          -881.0549577007893, CB=-1245.2679469540894,
                                          HF0=63442.20077641379, VF0=-27995.71383270186, Tol=1e-06, MaxIter=50, plots=2)


    #(fAH, fAV, fBH, fBV, info) = catenary(89.9, 59.2, 130.0, 751000000.0,
    #                                      881.05, CB=-372.7, Tol=1e-06, MaxIter=50, plots=2)

    #(fAH, fAV, fBH, fBV, info) = catenary(400, 200, 500.0, 7510000000000.0, 200.0, CB=-372.7, Tol=1e-06, MaxIter=50, plots=3)
    #
    """

    """
    #(fAH, fAV, fBH, fBV, info) = catenary(400, 200, 500.0, 7510000000000.0, 200.0, CB=5.0, Tol=1e-06, MaxIter=50, plots=3)
    (fAH, fAV, fBH, fBV, info) = catenary(400, 200, 500.0, 7510000000000.0, 200.0, CB=-20, Tol=1e-06, MaxIter=50, plots=3)


    print(f"Error is {info['catenary']['err'][0]:8.3f}, {info['catenary']['err'][1]:8.3f} m")

    print(" Fax={:8.2e}, Faz={:8.2e}, Fbx={:8.2e}, Fbz={:8.2e}".format(fAH, fAV, fBH, fBV))
    print(info['jacobian'])
    print(np.linalg.inv(info['jacobian']))
    """

    # (fAH, fAV, fBH, fBV, info) = catenary(100, 20, 130, 1e12, 100.0, CB=-20, Tol=0.001, MaxIter=50, plots=3)
    # (fAH, fAV, fBH, fBV, info) = catenary(205, -3.9, 250, 1229760000.0, 2442, CB=-55, Tol=1e-06, MaxIter=50, plots=3)

    (fAH1, fAV1, fBH1, fBV1, info1) = catenary(400, 100, 470, 1e12, 100.0, CB=-10, Tol=0.001, MaxIter=50, plots=4)
    # (fAH1, fAV1, fBH1, fBV1, info1) = catenary(2306.4589923684835, 1.225862496312402e-05, 1870.0799339749528, 203916714.02425563, 405.04331583394304, CB=0.0, HF0=58487689.78903873, VF0=0.0, Tol=4.000000000000001e-06, MaxIter=50, plots=1)
    # (fAH1, fAV1, fBH1, fBV1, info1) = catenary(459.16880261639346, 0.0004792939078015479, 447.67890341877506, 2533432597.6567926, 5032.201233267459, CB=0.0, HF0=65021800.32966018, VF0=17487.252675845888, Tol=4.000000000000001e-06, MaxIter=50, plots=1)
    # (fAH1, fAV1, fBH1, fBV1, info1) = catenary(0.00040612281105723014, 391.558570722038, 400.0, 3300142385.3140063, 6555.130220040344, CB=-287.441429277962, HF0=2127009.4122708915, VF0=10925834.69512347, Tol=4.000000000000001e-06, MaxIter=100, plots=1)
    # (fAH1, fAV1, fBH1, fBV1, info1) = catenary(0.0004959907076624859, 69.87150531147275, 110.89397565668423, 80297543.26800226, 146.26820268238743, CB=-509.12849468852727, HF0=1322712.3676957292, VF0=1045583.1849462093, Tol=4.000000000000001e-06, MaxIter=50, plots=1)

    plt.plot(info1["X"], info1["Z"])
    # plt.plot(infoU['X'], infoU['Z'] )
    plt.axis("equal")
    """
    plt.figure()
    plt.plot(info1['s'], info1['Te'] )
    plt.plot(infoU['s'], infoU['Te'] )
    """

    print(fAH1, fAV1, fBH1, fBV1)
    print("")

    from wisdem.moorpy.helpers import printMat

    printMat(info1["K"])
    print("")
    printMat(info1["stiffnessA"])
    print("")
    printMat(info1["stiffnessB"])

    plt.show()



import unittest

import numpy as np
import openmdao.api as om
import wisdem.drivetrainse.hub as hub


class TestHub(unittest.TestCase):
    def testMaxTorque(self):
        torq = hub.FindMaxTorque()
        inputs = {}
        outputs = {}

        inputs["blades_I"] = np.ones(6)
        inputs["blades_I"][0] = 1e3
        inputs["rated_rpm"] = 10.0
        inputs["stop_time"] = 5.0

        torq.compute(inputs, outputs)
        self.assertAlmostEqual(outputs["max_torque"], 1e4 * np.pi / 30.0 / 5.0)

    def testPitch(self):

        pitch = hub.PitchSystem()
        inputs = {}
        outputs = {}
        discrete_inputs = {}
        discrete_outputs = {}

        inputs["blade_mass"] = 17740.0  # kg
        discrete_inputs["n_blades"] = 3
        inputs["hub_diameter"] = 4.0
        inputs["rho"] = 7850.0
        inputs["Xy"] = 250e6

        AirDensity = 1.225  # kg/(m^3)
        Solidity = 0.0517
        RatedWindSpeed = 11.05  # m/s
        rotor_diameter = 126.0
        inputs["BRFM"] = (
            (3.06 * np.pi / 8)
            * AirDensity
            * (RatedWindSpeed ** 2)
            * (Solidity * (rotor_diameter ** 3))
            / discrete_inputs["n_blades"]
        )

        inputs["pitch_system_scaling_factor"] = 1.0

        pitch.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        m = outputs["pitch_mass"]
        self.assertAlmostEqual(outputs["pitch_I"][0], m * 4)
        self.assertAlmostEqual(outputs["pitch_I"][1], m * 2)
        self.assertAlmostEqual(outputs["pitch_I"][2], m * 2)

    def testRegression(self):
        opt = {}
        opt["hub_gamma"] = 2.0
        opt["spinner_gamma"] = 1.5
        hub_prob = om.Problem(model=hub.Hub_System(modeling_options=opt))
        hub_prob.setup()

        hub_prob["blades_I"] = 1e3 * np.ones(6)
        hub_prob["blades_I"][0] = 199200777.51 * 30.0 * 5 / 10.0 / np.pi
        hub_prob["rated_rpm"] = 10.0
        hub_prob["stop_time"] = 5.0
        hub_prob["n_blades"] = 3
        hub_prob["flange_t2shell_t"] = 4.0
        hub_prob["flange_OD2hub_D"] = 0.5
        hub_prob["flange_ID2flange_OD"] = 0.8
        hub_prob["hub_in2out_circ"] = 1.2
        hub_prob["hub_stress_concentration"] = 2.5
        hub_prob["n_front_brackets"] = 3
        hub_prob["n_rear_brackets"] = 3
        hub_prob["clearance_hub_spinner"] = 0.5
        hub_prob["spin_hole_incr"] = 1.2
        hub_prob["blade_root_diameter"] = 4.5
        hub_prob["hub_diameter"] = 6.235382907247958

        hub_prob["pitch_system.blade_mass"] = 17000.0
        hub_prob["pitch_system.BRFM"] = 1.0e6
        hub_prob["pitch_system_scaling_factor"] = 0.54
        hub_prob["pitch_system.rho"] = 7850.0
        hub_prob["pitch_system.Xy"] = 371.0e6

        hub_prob["hub_shell.rho"] = 7200.0
        # hub_prob['hub_shell.max_torque']           = 199200777.51
        hub_prob["hub_shell.Xy"] = 200.0e6
        hub_prob["hub_shell.metal_cost"] = 3.00

        hub_prob["spinner_gust_ws"] = 70
        hub_prob["spinner.composite_Xt"] = 60.0e6
        # hub_prob['spinner.composite_SF']           = 1.5
        hub_prob["spinner.composite_rho"] = 1600.0
        hub_prob["spinner.Xy"] = 225.0e6
        # hub_prob['spinner.metal_SF']               = 1.5
        hub_prob["spinner.metal_rho"] = 7850.0
        hub_prob["spinner.composite_cost"] = 7.00
        hub_prob["spinner.metal_cost"] = 3.00

        hub_prob.run_model()

        self.assertAlmostEqual(hub_prob["pitch_system.pitch_mass"][0], 6202.76603773585)
        self.assertAlmostEqual(hub_prob["hub_shell.hub_mass"][0], 79417.52737564275)
        self.assertAlmostEqual(hub_prob["hub_shell.hub_cost"][0], 238252.58212692826)
        self.assertAlmostEqual(hub_prob["hub_shell.hub_cm"][0], 2.8681137824204033)
        self.assertAlmostEqual(hub_prob["hub_shell.hub_I"][0], 514625.57739416)
        self.assertAlmostEqual(hub_prob["spinner.spinner_mass"][0], 1393.1009978176617)  # 1704.965737284796)
        self.assertAlmostEqual(hub_prob["spinner.spinner_diameter"][0], 7.235382907247958)
        self.assertAlmostEqual(hub_prob["spinner.spinner_cost"][0], 7678.042964192897)  # 9395.05078685929)
        self.assertAlmostEqual(hub_prob["spinner.spinner_cm"][0], 3.117691453623979)
        self.assertAlmostEqual(hub_prob["spinner.spinner_I"][0], 12154.984015448832)  # 14876.04367239)
        self.assertAlmostEqual(hub_prob["adder.hub_system_mass"][0], 87013.39441119626)  # 87325.2591506634)
        self.assertAlmostEqual(hub_prob["adder.hub_system_cost"][0], 245930.62509112115)  # 247647.63291378756)
        self.assertAlmostEqual(hub_prob["adder.hub_system_cm"][0], 2.872109568415645)  # 2.8729866151188928)
        self.assertAlmostEqual(hub_prob["adder.hub_system_I"][0], 587071.4472964063)  # 589792.50695335)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestHub))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2021, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from wisdem.orbit.phases.design import DesignPhase


class GbfDesign(DesignPhase):
    """Gravity Based Foundation Design Module"""

    expected_config = {
        "site": {"depth": "m"},
        "plant": {"num_turbines": "int"},
        "turbine": {"turbine_rating": "MW"},
        "gbf_design": {},
    }

    output_config = {
        "gbf": {  # TODO: Does this need to be 'substructure' to work with downstream modules?
            "height": "m",
            "mass": "t",
            "unit_cost": "USD",
        },
        "transition_piece": {
            "deck_space": "m2",
            "mass": "t",
            "unit_cost": "USD",
        },
    }

    def __init__(self, config, **kwargs):
        """Creates an instance of `GbfDesign`."""

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)
        self.gbf_design = self.config.get("gbf_design", {})

        self._outputs = {}

    def run(self):
        """Main run function used by ProjectManager."""

        self.design_gbf()

    def design_gbf(self):
        """Main GBF design script."""

        raise NotImplementedError("'GbfDesign' module has not been fullyimplemented yet.")

    @property
    def detailed_output(self):
        """Returns detailed output dictionary."""

        return {}

    @property
    def total_cost(self):
        """Returns total cost of subcomponent(s)."""

        num_turbines = self.config["plant"]["num_turbines"]
        return (self._outputs["gbf"]["unit_cost"] + self._outputs["transition_piece"]["unit_cost"]) * num_turbines

    @property
    def design_result(self):
        """Must match `self.output_config` structure."""

        return self._outputs



import numpy as np
from scipy.linalg import solve_banded
from wisdem.commonse.utilities import _checkIfFloat


class NaturalCubicSpline(object):
    """
    class implementation of utilities.cubic_with_deriv
    """

    def __init__(self, xp, yp):

        if np.any(np.diff(xp) < 0):
            raise TypeError("xp must be in ascending order")

        # n = len(x)
        self.m = len(xp)

        xk = xp[1:-1]
        yk = yp[1:-1]
        xkp = xp[2:]
        ykp = yp[2:]
        xkm = xp[:-2]
        ykm = yp[:-2]

        b = (ykp - yk) / (xkp - xk) - (yk - ykm) / (xk - xkm)
        l = (xk - xkm) / 6.0
        d = (xkp - xkm) / 3.0
        u = (xkp - xk) / 6.0
        # u[0] = 0.0  # non-existent entries
        # l[-1] = 0.0

        # solve for second derivatives
        fpp = solve_banded((1, 1), np.array([u, d, l]), b)
        self.fpp = np.concatenate([[0.0], fpp, [0.0]])  # natural spline
        self.xp = xp
        self.yp = yp

    def __call__(self, x, deriv=False):

        x, n = _checkIfFloat(x)
        y = np.zeros(n)
        dydx = np.zeros(n)
        dydxp = np.zeros((n, self.m))
        dydyp = np.zeros((n, self.m))

        # find location in vector
        for i in range(n):
            if x[i] < self.xp[0]:
                j = 0
            elif x[i] > self.xp[-1]:
                j = self.m - 2
            else:
                for j in range(self.m - 1):
                    if self.xp[j + 1] > x[i]:
                        break
            x1 = self.xp[j]
            y1 = self.yp[j]
            x2 = self.xp[j + 1]
            y2 = self.yp[j + 1]

            A = (x2 - x[i]) / (x2 - x1)
            B = 1 - A
            C = 1.0 / 6 * (A ** 3 - A) * (x2 - x1) ** 2
            D = 1.0 / 6 * (B ** 3 - B) * (x2 - x1) ** 2

            y[i] = A * y1 + B * y2 + C * self.fpp[j] + D * self.fpp[j + 1]
            dAdx = -1.0 / (x2 - x1)
            dBdx = -dAdx
            dCdx = 1.0 / 6 * (3 * A ** 2 - 1) * dAdx * (x2 - x1) ** 2
            dDdx = 1.0 / 6 * (3 * B ** 2 - 1) * dBdx * (x2 - x1) ** 2
            dydx[i] = dAdx * y1 + dBdx * y2 + dCdx * self.fpp[j] + dDdx * self.fpp[j + 1]

        if n == 1:
            y = y[0]
            dydx = dydx[0]

        if deriv:
            return y, dydx

        else:
            return y



import unittest

import numpy as np
import numpy.testing as npt
from wisdem.commonse.csystem import DirectionVector


class TestCSystem(unittest.TestCase):
    def test(self):

        x = np.array([1.0, 2.0])
        y = np.array([1.3, 4.3])
        z = np.array([2.3, 2.3])
        a = DirectionVector(x, y, z)

        x = np.array([3.2, 1.5])
        y = np.array([2.1, 3.2])
        z = np.array([5.6, 7.7])
        b = DirectionVector(x, y, z)

        c = a.cross(b)
        dx, dy, dz = a.cross_deriv(b)
        dx, dy, dz = a.cross_deriv_array(b)

        d = -a
        e = a + b
        f = a - b
        f += a
        f -= a
        k = a * b
        p = a / b
        p *= b
        p /= a

        e = a + 1.2
        f = a - 1.2
        f += 2.0
        f -= 3.0
        k = a * 4.0
        p = a / 2.0
        p *= 3.0
        p /= 2.0

        T1 = 5.0
        T2 = 4.0
        T3 = -3.0
        tilt = 20.0

        F = DirectionVector(T1, T2, T3).hubToYaw(tilt)

        Fp1 = DirectionVector(T1 + 1e-6, T2, T3).hubToYaw(tilt)
        Fp2 = DirectionVector(T1, T2 + 1e-6, T3).hubToYaw(tilt)
        Fp3 = DirectionVector(T1, T2, T3 + 1e-6).hubToYaw(tilt)
        Fp4 = DirectionVector(T1, T2, T3).hubToYaw(tilt + 1e-6)

        dFx = F.dx
        dFy = F.dy
        dFz = F.dz

        npt.assert_almost_equal(dFx["dx"], (Fp1.x - F.x) / 1e-6)
        npt.assert_almost_equal(dFx["dy"], (Fp2.x - F.x) / 1e-6)
        npt.assert_almost_equal(dFx["dz"], (Fp3.x - F.x) / 1e-6)
        npt.assert_almost_equal(dFx["dtilt"], (Fp4.x - F.x) / 1e-6)

        npt.assert_almost_equal(dFy["dx"], (Fp1.y - F.y) / 1e-6)
        npt.assert_almost_equal(dFy["dy"], (Fp2.y - F.y) / 1e-6)
        npt.assert_almost_equal(dFy["dz"], (Fp3.y - F.y) / 1e-6)
        npt.assert_almost_equal(dFy["dtilt"], (Fp4.y - F.y) / 1e-6)

        npt.assert_almost_equal(dFz["dx"], (Fp1.z - F.z) / 1e-6)
        npt.assert_almost_equal(dFz["dy"], (Fp2.z - F.z) / 1e-6)
        npt.assert_almost_equal(dFz["dz"], (Fp3.z - F.z) / 1e-6)
        npt.assert_almost_equal(dFz["dtilt"], (Fp4.z - F.z) / 1e-6)

        yaw = 3.0
        F = DirectionVector(T1, T2, T3).hubToYaw(tilt).yawToWind(yaw)

        Fp1 = DirectionVector(T1 + 1e-6, T2, T3).hubToYaw(tilt).yawToWind(yaw)
        Fp2 = DirectionVector(T1, T2 + 1e-6, T3).hubToYaw(tilt).yawToWind(yaw)
        Fp3 = DirectionVector(T1, T2, T3 + 1e-6).hubToYaw(tilt).yawToWind(yaw)
        Fp4 = DirectionVector(T1, T2, T3).hubToYaw(tilt + 1e-6).yawToWind(yaw)
        Fp5 = DirectionVector(T1, T2, T3).hubToYaw(tilt).yawToWind(yaw + 1e-6)

        dFx = F.dx
        dFy = F.dy
        dFz = F.dz

        npt.assert_almost_equal(dFx["dx"], (Fp1.x - F.x) / 1e-6)
        npt.assert_almost_equal(dFx["dy"], (Fp2.x - F.x) / 1e-6)
        npt.assert_almost_equal(dFx["dz"], (Fp3.x - F.x) / 1e-6)
        npt.assert_almost_equal(dFx["dtilt"], (Fp4.x - F.x) / 1e-6)
        npt.assert_almost_equal(dFx["dyaw"], (Fp5.x - F.x) / 1e-6)

        npt.assert_almost_equal(dFy["dx"], (Fp1.y - F.y) / 1e-6)
        npt.assert_almost_equal(dFy["dy"], (Fp2.y - F.y) / 1e-6)
        npt.assert_almost_equal(dFy["dz"], (Fp3.y - F.y) / 1e-6)
        npt.assert_almost_equal(dFy["dtilt"], (Fp4.y - F.y) / 1e-6)
        npt.assert_almost_equal(dFy["dyaw"], (Fp5.y - F.y) / 1e-6)

        npt.assert_almost_equal(dFz["dx"], (Fp1.z - F.z) / 1e-6)
        npt.assert_almost_equal(dFz["dy"], (Fp2.z - F.z) / 1e-6)
        npt.assert_almost_equal(dFz["dz"], (Fp3.z - F.z) / 1e-6)
        npt.assert_almost_equal(dFz["dtilt"], (Fp4.z - F.z) / 1e-6)
        npt.assert_almost_equal(dFz["dyaw"], (Fp5.z - F.z) / 1e-6)

        inertial_F = F.windToInertial(20.0)
        new_F = inertial_F.inertialToWind(20.0)
        npt.assert_almost_equal(F.toArray(), new_F.toArray())

        wind_F = F.yawToWind(15.0)
        new_F = wind_F.windToYaw(15.0)
        npt.assert_almost_equal(F.toArray(), new_F.toArray())

        yaw_F = F.hubToYaw(12.0)
        new_F = yaw_F.yawToHub(12.0)
        npt.assert_almost_equal(F.toArray(), new_F.toArray())

        az_F = F.hubToAzimuth(12.0)
        new_F = az_F.azimuthToHub(12.0)
        npt.assert_almost_equal(F.toArray(), new_F.toArray())

        blade_F = F.azimuthToBlade(12.0)
        new_F = blade_F.bladeToAzimuth(12.0)
        npt.assert_almost_equal(F.toArray(), new_F.toArray())

        blade_F = F.airfoilToBlade(12.0)
        new_F = blade_F.bladeToAirfoil(12.0)
        npt.assert_almost_equal(F.toArray(), new_F.toArray())

        profile_F = F.airfoilToProfile()
        new_F = profile_F.profileToAirfoil()
        npt.assert_almost_equal(F.toArray(), new_F.toArray())


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestCSystem))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



import copy
import unittest

import numpy as np
import openmdao.api as om
import numpy.testing as npt

from wisdem.commonse import gravity as g
from wisdem.fixed_bottomse.jacket import JacketSE


class Test(unittest.TestCase):
    def testAll(self):
        modeling_options = {}
        modeling_options["WISDEM"] = {}
        modeling_options["WISDEM"]["n_dlc"] = 1
        modeling_options["WISDEM"]["FixedBottomSE"] = {}
        modeling_options["WISDEM"]["FixedBottomSE"]["mud_brace"] = True
        modeling_options["WISDEM"]["FixedBottomSE"]["n_legs"] = 3
        modeling_options["WISDEM"]["FixedBottomSE"]["n_bays"] = 2
        modeling_options["WISDEM"]["FixedBottomSE"]["gamma_f"] = 1.0
        modeling_options["WISDEM"]["FixedBottomSE"]["gamma_m"] = 1.0
        modeling_options["WISDEM"]["FixedBottomSE"]["gamma_n"] = 1.0
        modeling_options["WISDEM"]["FixedBottomSE"]["gamma_b"] = 1.0
        modeling_options["WISDEM"]["FixedBottomSE"]["material"] = "steel"
        modeling_options["WISDEM"]["FixedBottomSE"]["frame3dd"] = {}
        modeling_options["WISDEM"]["FixedBottomSE"]["frame3dd"]["shear"] = 1
        modeling_options["WISDEM"]["FixedBottomSE"]["frame3dd"]["geom"] = 1
        modeling_options["WISDEM"]["FixedBottomSE"]["save_truss_figures"] = False
        modeling_options["materials"] = {}
        modeling_options["materials"]["n_mat"] = 1

        prob = om.Problem(
            model=JacketSE(
                modeling_options=modeling_options,
            )
        )

        prob.setup()

        prob["material_names"] = ["steel"]
        prob["E_mat"] = 1e9 * np.ones((1, 3))
        prob["G_mat"] = 1e8 * np.ones((1, 3))
        prob["rho_mat"] = 1e4
        prob["sigma_y_mat"] = 1e8

        prob.run_model()


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(Test))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



"""
Copyright (c) NREL. All rights reserved.
"""

from __future__ import print_function
import numpy as np

import openmdao.api as om

from wisdem.nrelcsm.nrel_csm_cost_2015 import Turbine_CostsSE_2015


# --------------------------------------------------------------------
class BladeMass(om.ExplicitComponent):
    """
    Compute blade mass of the form :math:`mass = k*diameter^b`.
    Value of :math:`k` was updated in 2015 to be 0.5.
    Value of :math:`b` was updated to be 2.47/2.54 for turbine class I blades with/without carbon or
    2.44/2.5 for other turbine classes with/without carbon.
    Values of k and b can be overridden by the user with use of `blade_mass_coeff` (k) and/or `blade_user_exp` (b).
    To use `blade_user_exp`, the value of `turbine_class` must be less than 1.

    Parameters
    ----------
    rotor_diameter : float, [m]
        rotor diameter of the machine
    turbine_class : float
        turbine class.  Set to 1 for Class I, 2 for Class II+, or 0 for user overrides of blade_user_exp
    blade_has_carbon : boolean
        does the blade have carbon?
    blade_mass_coeff : float
        k in the blade mass equation: k*(rotor_diameter/2)^b
    blade_user_exp : float
        optional user-entered exp for the blade mass equation

    Returns
    -------
    blade_mass : float, [kg]
        component mass

    """

    def setup(self):
        self.add_input("rotor_diameter", 0.0, units="m")
        self.add_discrete_input("turbine_class", 1)
        self.add_discrete_input("blade_has_carbon", False)
        self.add_input("blade_mass_coeff", 0.5)
        self.add_input("blade_user_exp", 2.5)

        self.add_output("blade_mass", 0.0, units="kg")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        rotor_diameter = inputs["rotor_diameter"]
        turbine_class = discrete_inputs["turbine_class"]
        blade_has_carbon = discrete_inputs["blade_has_carbon"]
        blade_mass_coeff = inputs["blade_mass_coeff"]
        blade_user_exp = inputs["blade_user_exp"]

        # select the exp for the blade mass equation
        exp = 0.0
        if turbine_class == 1:
            if blade_has_carbon:
                exp = 2.47
            else:
                exp = 2.54
        elif turbine_class > 1:
            if blade_has_carbon:
                exp = 2.44
            else:
                exp = 2.50
        else:
            exp = blade_user_exp

        # calculate the blade mass
        outputs["blade_mass"] = blade_mass_coeff * (rotor_diameter / 2) ** exp


# --------------------------------------------------------------------
class HubMass(om.ExplicitComponent):
    """
    Compute hub mass in the form of :math:`mass = k*m_{blade} + b`.
    Value of :math:`k` was updated in 2015 to be 2.3.
    Value of :math:`b` was updated in 2015 to be 1320.

    Parameters
    ----------
    blade_mass : float, [kg]
        component mass
    hub_mass_coeff : float
        k inthe hub mass equation: k*blade_mass + b
    hub_mass_intercept : float
        b in the hub mass equation: k*blade_mass + b

    Returns
    -------
    hub_mass : float, [kg]
        component mass

    """

    def setup(self):
        # Variables
        self.add_input("blade_mass", 0.0, units="kg")
        self.add_input("hub_mass_coeff", 2.3)
        self.add_input("hub_mass_intercept", 1320.0)

        self.add_output("hub_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):
        blade_mass = inputs["blade_mass"]
        hub_mass_coeff = inputs["hub_mass_coeff"]
        hub_mass_intercept = inputs["hub_mass_intercept"]

        # calculate the hub mass
        outputs["hub_mass"] = hub_mass_coeff * blade_mass + hub_mass_intercept


# --------------------------------------------------------------------
class PitchSystemMass(om.ExplicitComponent):
    """
    Compute pitch bearing mass in the form of :math:`m_{bearing} = k*m_{blade}*nblade + b1`.
    Then compute pitch system mass, with bearing housing in the form of :math:`mass = (1+h)*m_{bearing} + b2`.
    The values of the constants were NOT updated in 2015 and are the same as the original CSM.
    Value of :math:`k` is 0.1295.
    Value of :math:`h` is 0.328.
    Value of :math:`b1` is 491.31.
    Value of :math:`b2` is 555.0.

    Parameters
    ----------
    blade_mass : float, [kg]
        component mass
    blade_number : float
        number of rotor blades
    pitch_bearing_mass_coeff : float
        k in the pitch bearing mass equation: k*blade_mass*blade_number + b
    pitch_bearing_mass_intercept : float
        b in the pitch bearing mass equation: k*blade_mass*blade_number + b
    bearing_housing_fraction : float
        bearing housing fraction
    mass_sys_offset : float
        mass system offset

    Returns
    -------
    pitch_system_mass : float, [kg]
        component mass

    """

    def setup(self):
        self.add_input("blade_mass", 0.0, units="kg")
        self.add_discrete_input("blade_number", 3)
        self.add_input("pitch_bearing_mass_coeff", 0.1295)
        self.add_input("pitch_bearing_mass_intercept", 491.31)
        self.add_input("bearing_housing_fraction", 0.3280)
        self.add_input("mass_sys_offset", 555.0)

        self.add_output("pitch_system_mass", 0.0, units="kg")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        blade_mass = inputs["blade_mass"]
        blade_number = discrete_inputs["blade_number"]
        pitch_bearing_mass_coeff = inputs["pitch_bearing_mass_coeff"]
        pitch_bearing_mass_intercept = inputs["pitch_bearing_mass_intercept"]
        bearing_housing_fraction = inputs["bearing_housing_fraction"]
        mass_sys_offset = inputs["mass_sys_offset"]

        # calculate the hub mass
        pitchBearingMass = pitch_bearing_mass_coeff * blade_mass * blade_number + pitch_bearing_mass_intercept
        outputs["pitch_system_mass"] = pitchBearingMass * (1 + bearing_housing_fraction) + mass_sys_offset


# --------------------------------------------------------------------
class SpinnerMass(om.ExplicitComponent):
    """
    Compute spinner (nose cone) mass in the form of :math:`mass = k*diameter + b`.
    Value of :math:`k` was updated in 2015 to be 15.5.
    Value of :math:`b` was updated in 2015 to be -980.

    Parameters
    ----------
    rotor_diameter : float, [m]
        rotor diameter of the machine
    spinner_mass_coeff : float
        k inthe spinner mass equation: k*rotor_diameter + b
    spinner_mass_intercept : float
        b in the spinner mass equation: k*rotor_diameter + b

    Returns
    -------
    spinner_mass : float, [kg]
        component mass

    """

    def setup(self):
        self.add_input("rotor_diameter", 0.0, units="m")
        self.add_input("spinner_mass_coeff", 15.5)
        self.add_input("spinner_mass_intercept", -980.0)

        self.add_output("spinner_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):
        rotor_diameter = inputs["rotor_diameter"]
        spinner_mass_coeff = inputs["spinner_mass_coeff"]
        spinner_mass_intercept = inputs["spinner_mass_intercept"]

        # calculate the spinner mass
        outputs["spinner_mass"] = spinner_mass_coeff * rotor_diameter + spinner_mass_intercept


# --------------------------------------------------------------------
class LowSpeedShaftMass(om.ExplicitComponent):
    """
    Compute low speed shaft mass in the form of :math:`mass = k*(m_{blade}*power)^b1 + b2`.
    Value of :math:`k` was updated in 2015 to be 13.
    Value of :math:`b1` was updated in 2015 to be 0.65.
    Value of :math:`b2` was updated in 2015 to be 775.

    Parameters
    ----------
    blade_mass : float, [kg]
        mass for a single wind turbine blade
    machine_rating : float, [kW]
        machine rating
    lss_mass_coeff : float
        k inthe lss mass equation: k*(blade_mass*rated_power)^b1 + b2
    lss_mass_exp : float
        b1 in the lss mass equation: k*(blade_mass*rated_power)^b1 + b2
    lss_mass_intercept : float
        b2 in the lss mass equation: k*(blade_mass*rated_power)^b1 + b2

    Returns
    -------
    lss_mass : float, [kg]
        component mass

    """

    def setup(self):
        self.add_input("blade_mass", 0.0, units="kg")
        self.add_input("machine_rating", 0.0, units="kW")
        self.add_input("lss_mass_coeff", 13.0)
        self.add_input("lss_mass_exp", 0.65)
        self.add_input("lss_mass_intercept", 775.0)

        self.add_output("lss_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):
        blade_mass = inputs["blade_mass"]
        machine_rating = inputs["machine_rating"]
        lss_mass_coeff = inputs["lss_mass_coeff"]
        lss_mass_exp = inputs["lss_mass_exp"]
        lss_mass_intercept = inputs["lss_mass_intercept"]

        # calculate the lss mass
        outputs["lss_mass"] = (
            lss_mass_coeff * (blade_mass * machine_rating / 1000.0) ** lss_mass_exp + lss_mass_intercept
        )


# --------------------------------------------------------------------
class BearingMass(om.ExplicitComponent):
    """
    Compute main bearing mass (single bearing) in the form of :math:`mass = k*diameter^b`.
    Value of :math:`k` was updated in 2015 to be 1e-4.
    Value of :math:`b` was updated in 2015 to be 3.5.

    Parameters
    ----------
    rotor_diameter : float, [m]
        rotor diameter of the machine
    bearing_mass_coeff : float
        k inthe bearing mass equation: k*rotor_diameter^b
    bearing_mass_exp : float
        exp in the bearing mass equation: k*rotor_diameter^b

    Returns
    -------
    main_bearing_mass : float, [kg]
        component mass

    """

    def setup(self):
        # Variables
        self.add_input("rotor_diameter", 0.0, units="m")
        self.add_input("bearing_mass_coeff", 0.0001)
        self.add_input("bearing_mass_exp", 3.5)

        self.add_output("main_bearing_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):
        rotor_diameter = inputs["rotor_diameter"]
        bearing_mass_coeff = inputs["bearing_mass_coeff"]
        bearing_mass_exp = inputs["bearing_mass_exp"]

        # calculates the mass of a SINGLE bearing
        outputs["main_bearing_mass"] = bearing_mass_coeff * rotor_diameter ** bearing_mass_exp


# --------------------------------------------------------------------
class RotorTorque(om.ExplicitComponent):
    """
    Computed rated rpm and rotor torque from rated power, rotor diameter, max tip speed, and drivetrain efficiency.
    Rotor torque will be used to size other drivetrain components, such as the generator.

    Parameters
    ----------
    rotor_diameter : float, [m]
        rotor diameter of the machine
    machine_rating : float, [kW]
        machine rating
    max_tip_speed : float, [m/s]
        Maximum allowable blade tip speed
    max_efficiency : float
        Maximum possible drivetrain efficiency

    Returns
    -------
    rated_rpm : float, [rpm]
        rpm of rotor at rated power
    rotor_torque : float, [N*m]
        torque from rotor at rated power

    """

    def setup(self):
        self.add_input("rotor_diameter", 0.0, units="m")
        self.add_input("machine_rating", 0.0, units="kW")
        self.add_input("max_tip_speed", 0.0, units="m/s")
        self.add_input("max_efficiency", 0.0)

        self.add_output("rated_rpm", 0.0, units="rpm")
        self.add_output("rotor_torque", 0.0, units="N*m")

    def compute(self, inputs, outputs):
        # Rotor force calculations for nacelle inputs
        maxTipSpd = inputs["max_tip_speed"]
        maxEfficiency = inputs["max_efficiency"]

        ratedHubPower_W = inputs["machine_rating"] * 1000.0 / maxEfficiency
        rotorSpeed = maxTipSpd / (0.5 * inputs["rotor_diameter"])
        outputs["rated_rpm"] = rotorSpeed / (2 * np.pi) * 60.0
        outputs["rotor_torque"] = ratedHubPower_W / rotorSpeed


# --------------------------------------------------------------------
class GearboxMass(om.ExplicitComponent):
    """
    Compute gearbox mass in the form of :math:`mass = k*torque^b`.
    Value of :math:`k` was updated in 2015 to be 113.
    Value of :math:`b` was updated in 2015 to be 0.71.

    Parameters
    ----------
    rotor_torque : float, [N*m]
        torque from rotor at rated power
    gearbox_mass_coeff : float
        k inthe gearbox mass equation: k*rotor_torque^b
    gearbox_mass_exp : float
        exp in the gearbox mass equation: k*rotor_torque^b

    Returns
    -------
    gearbox_mass : float, [kg]
        component mass

    """

    def setup(self):
        self.add_input("rotor_torque", 0.0, units="N*m")
        self.add_input("gearbox_mass_coeff", 113.0)
        self.add_input("gearbox_mass_exp", 0.71)

        self.add_output("gearbox_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):
        rotor_torque = inputs["rotor_torque"]
        gearbox_mass_coeff = inputs["gearbox_mass_coeff"]
        gearbox_mass_exp = inputs["gearbox_mass_exp"]

        # calculate the gearbox mass
        outputs["gearbox_mass"] = gearbox_mass_coeff * (rotor_torque / 1000.0) ** gearbox_mass_exp


# --------------------------------------------------------------------
class BrakeMass(om.ExplicitComponent):
    """
    Compute brake mass in the form of :math:`mass = k*torque`.
    Value of :math:`k` was updated in 2020 to be 0.00122.

    Parameters
    ----------
    rotor_torque : float, [N*m]
        rotor torque at rated power
    brake_mass_coeff : float
        Mass scaling coefficient

    Returns
    -------
    brake_mass : float, [kg]
        overall component mass
    """

    def setup(self):
        self.add_input("rotor_torque", 0.0, units="N*m")
        self.add_input("brake_mass_coeff", 0.00122)

        self.add_output("brake_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):

        # Unpack inputs
        rotor_torque = float(inputs["rotor_torque"])
        coeff = float(inputs["brake_mass_coeff"])

        # Regression based sizing derived by J.Keller under FOA 1981 support project
        outputs["brake_mass"] = coeff * rotor_torque


# --------------------------------------------------------------------
class HighSpeedShaftMass(om.ExplicitComponent):
    """
    Compute high speed shaft mass in the form of :math:`mass = k*power`.
    Value of :math:`k` was updated in 2015 to be 0.19894.

    Parameters
    ----------
    machine_rating : float, [kW]
        machine rating
    hss_mass_coeff : float
        NREL CSM hss equation; removing intercept since it is negligible

    Returns
    -------
    hss_mass : float, [kg]
        component mass

    """

    def setup(self):
        self.add_input("machine_rating", 0.0, units="kW")
        self.add_input("hss_mass_coeff", 0.19894)

        self.add_output("hss_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):
        machine_rating = inputs["machine_rating"]
        hss_mass_coeff = inputs["hss_mass_coeff"]

        outputs["hss_mass"] = hss_mass_coeff * machine_rating


# --------------------------------------------------------------------
class GeneratorMass(om.ExplicitComponent):
    """
    Compute generator mass in the form of :math:`mass = k*power + b`.
    Value of :math:`k` was updated in 2015 to be 2300.
    Value of :math:`b` was updated in 2015 to be 3400.

    Parameters
    ----------
    machine_rating : float, [kW]
        machine rating
    generator_mass_coeff : float
        k inthe generator mass equation: k*rated_power + b
    generator_mass_intercept : float
        b in the generator mass equation: k*rated_power + b

    Returns
    -------
    generator_mass : float, [kg]
        component mass

    """

    def setup(self):
        self.add_input("machine_rating", 0.0, units="kW")
        self.add_input("generator_mass_coeff", 2300.0)
        self.add_input("generator_mass_intercept", 3400.0)

        self.add_output("generator_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):
        machine_rating = inputs["machine_rating"]
        generator_mass_coeff = inputs["generator_mass_coeff"]
        generator_mass_intercept = inputs["generator_mass_intercept"]

        # calculate the generator mass
        outputs["generator_mass"] = generator_mass_coeff * machine_rating / 1000.0 + generator_mass_intercept


# --------------------------------------------------------------------
class BedplateMass(om.ExplicitComponent):
    """
    Compute bedplate mass in the form of :math:`mass = diameter^b`.
    Value of :math:`b` was updated in 2015 to be 2.2.

    Parameters
    ----------
    rotor_diameter : float, [m]
        rotor diameter of the machine
    bedplate_mass_exp : float
        exp in the bedplate mass equation: rotor_diameter^b

    Returns
    -------
    bedplate_mass : float, [kg]
        component mass

    """

    def setup(self):
        self.add_input("rotor_diameter", 0.0, units="m")
        self.add_input("bedplate_mass_exp", 2.2)

        self.add_output("bedplate_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):
        rotor_diameter = inputs["rotor_diameter"]
        bedplate_mass_exp = inputs["bedplate_mass_exp"]

        # calculate the bedplate mass
        outputs["bedplate_mass"] = rotor_diameter ** bedplate_mass_exp


# --------------------------------------------------------------------
class YawSystemMass(om.ExplicitComponent):
    """
    Compute yaw system mass in the form of :math:`mass = k*diameter^b`.
    The values of the constants were NOT updated in 2015 and are the same as the original CSM.
    Value of :math:`k` is 9e-4.
    Value of :math:`b` is 3.314.

    Parameters
    ----------
    rotor_diameter : float, [m]
        rotor diameter of the machine
    yaw_mass_coeff : float
        k inthe yaw mass equation: k*rotor_diameter^b
    yaw_mass_exp : float
        exp in the yaw mass equation: k*rotor_diameter^b

    Returns
    -------
    yaw_mass : float, [kg]
        component mass

    """

    def setup(self):
        self.add_input("rotor_diameter", 0.0, units="m")
        self.add_input("yaw_mass_coeff", 0.0009)
        self.add_input("yaw_mass_exp", 3.314)

        self.add_output("yaw_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):
        rotor_diameter = inputs["rotor_diameter"]
        yaw_mass_coeff = inputs["yaw_mass_coeff"]
        yaw_mass_exp = inputs["yaw_mass_exp"]

        # calculate yaw system mass #TODO - 50% adder for non-bearing mass
        outputs["yaw_mass"] = 1.5 * (
            yaw_mass_coeff * rotor_diameter ** yaw_mass_exp
        )  # JMF do we really want to expose all these?


# TODO: no variable speed mass; ignore for now

# --------------------------------------------------------------------
class HydraulicCoolingMass(om.ExplicitComponent):
    """
    Compute hydraulic cooling mass in the form of :math:`mass = k*power`.
    The values of the constants were NOT updated in 2015 and are the same as the original CSM.
    Value of :math:`k` is 0.08.

    Parameters
    ----------
    machine_rating : float, [kW]
        machine rating
    hvac_mass_coeff : float
        hvac linear coeff

    Returns
    -------
    hvac_mass : float, [kg]
        component mass

    """

    def setup(self):
        self.add_input("machine_rating", 0.0, units="kW")
        self.add_input("hvac_mass_coeff", 0.08)

        self.add_output("hvac_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):
        machine_rating = inputs["machine_rating"]
        hvac_mass_coeff = inputs["hvac_mass_coeff"]

        # calculate hvac system mass
        outputs["hvac_mass"] = hvac_mass_coeff * machine_rating


# --------------------------------------------------------------------
class NacelleCoverMass(om.ExplicitComponent):
    """
    Compute nacelle cover mass in the form of :math:`mass = k*power + b`.
    The values of the constants were NOT updated in 2015 and are the same as the original CSM.
    Value of :math:`k` is 1.2817.
    Value of :math:`b` is 428.19.

    Parameters
    ----------
    machine_rating : float, [kW]
        machine rating
    cover_mass_coeff : float
        k inthe spinner mass equation: k*rotor_diameter + b
    cover_mass_intercept : float
        b in the spinner mass equation: k*rotor_diameter + b

    Returns
    -------
    cover_mass : float, [kg]
        component mass

    """

    def setup(self):
        self.add_input("machine_rating", 0.0, units="kW")
        self.add_input("cover_mass_coeff", 1.2817)
        self.add_input("cover_mass_intercept", 428.19)

        self.add_output("cover_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):
        machine_rating = inputs["machine_rating"]
        cover_mass_coeff = inputs["cover_mass_coeff"]
        cover_mass_intercept = inputs["cover_mass_intercept"]

        # calculate nacelle cover mass
        outputs["cover_mass"] = cover_mass_coeff * machine_rating + cover_mass_intercept


# TODO: ignoring controls and electronics mass for now

# --------------------------------------------------------------------
class PlatformsMainframeMass(om.ExplicitComponent):
    """
    Compute platforms mass in the form of :math:`mass = k*m_{bedplate}` and
    crane mass as 3000kg, if flagged by the user.
    The values of the constants were NOT updated in 2015 and are the same as the original CSM.
    Value of :math:`k` is 0.125.

    Parameters
    ----------
    bedplate_mass : float, [kg]
        component mass
    platforms_mass_coeff : float
        nacelle platforms mass coeff as a function of bedplate mass [kg/kg]
    crane : boolean
        flag for presence of onboard crane
    crane_weight : float, [kg]
        weight of onboard crane

    Returns
    -------
    platforms_mass : float, [kg]
        component mass

    """

    # nacelle platforms, service crane, base hardware

    def setup(self):
        self.add_input("bedplate_mass", 0.0, units="kg")
        self.add_input("platforms_mass_coeff", 0.125)
        self.add_discrete_input("crane", False)
        self.add_input("crane_weight", 3000.0, units="kg")

        self.add_output("platforms_mass", 0.0, units="kg")
        # TODO: there is no base hardware mass model in the old model. Cost is not dependent on mass.

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        bedplate_mass = inputs["bedplate_mass"]
        platforms_mass_coeff = inputs["platforms_mass_coeff"]
        crane = discrete_inputs["crane"]
        crane_weight = inputs["crane_weight"]

        # calculate nacelle cover mass
        platforms_mass = platforms_mass_coeff * bedplate_mass

        # --- crane ---
        if crane:
            crane_mass = crane_weight
        else:
            crane_mass = 0.0

        outputs["platforms_mass"] = platforms_mass + crane_mass


# --------------------------------------------------------------------
class TransformerMass(om.ExplicitComponent):
    """
    Compute transformer mass in the form of :math:`mass = k*power + b`.
    Value of :math:`k` was updated in 2015 to be 1915.
    Value of :math:`b` was updated in 2015 to be 1910.

    Parameters
    ----------
    machine_rating : float, [kW]
        machine rating
    transformer_mass_coeff : float
        k inthe transformer mass equation: k*rated_power + b
    transformer_mass_intercept : float
        b in the transformer mass equation: k*rated_power + b

    Returns
    -------
    transformer_mass : float, [kg]
        component mass

    """

    def setup(self):
        self.add_input("machine_rating", 0.0, units="kW")
        self.add_input("transformer_mass_coeff", 1915.0)
        self.add_input("transformer_mass_intercept", 1910.0)

        self.add_output("transformer_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):
        machine_rating = inputs["machine_rating"]
        transformer_mass_coeff = inputs["transformer_mass_coeff"]
        transformer_mass_intercept = inputs["transformer_mass_intercept"]

        # calculate the transformer mass
        outputs["transformer_mass"] = transformer_mass_coeff * machine_rating / 1000.0 + transformer_mass_intercept


# --------------------------------------------------------------------
class TowerMass(om.ExplicitComponent):
    """
    Compute tower mass in the form of :math:`mass = k*H_{hub}^b`.
    Value of :math:`k` was updated in 2015 to be 19.828.
    Value of :math:`b` was updated in 2015 to be 2.0282.

    Parameters
    ----------
    hub_height : float, [m]
        hub height of wind turbine above ground / sea level
    tower_mass_coeff : float
        k inthe tower mass equation: k*hub_height^b
    tower_mass_exp : float
        b in the tower mass equation: k*hub_height^b

    Returns
    -------
    tower_mass : float, [kg]
        component mass

    """

    def setup(self):
        self.add_input("hub_height", 0.0, units="m")
        self.add_input("tower_mass_coeff", 19.828)
        self.add_input("tower_mass_exp", 2.0282)

        self.add_output("tower_mass", 0.0, units="kg")

    def compute(self, inputs, outputs):
        hub_height = inputs["hub_height"]
        tower_mass_coeff = inputs["tower_mass_coeff"]
        tower_mass_exp = inputs["tower_mass_exp"]

        # calculate the tower mass
        outputs["tower_mass"] = tower_mass_coeff * hub_height ** tower_mass_exp


# Turbine mass adder
class TurbineMassAdder(om.ExplicitComponent):
    """
    Aggregates all components masses into category labels of hub system, rotor, nacelle, and tower.

    Parameters
    ----------
    blade_mass : float, [kg]
        component mass
    hub_mass : float, [kg]
        component mass
    pitch_system_mass : float, [kg]
        component mass
    spinner_mass : float, [kg]
        component mass
    lss_mass : float, [kg]
        component mass
    main_bearing_mass : float, [kg]
        component mass
    gearbox_mass : float, [kg]
        component mass
    hss_mass : float, [kg]
        component mass
    brake_mass : float, [kg]
        component mass
    generator_mass : float, [kg]
        component mass
    bedplate_mass : float, [kg]
        component mass
    yaw_mass : float, [kg]
        component mass
    hvac_mass : float, [kg]
        component mass
    cover_mass : float, [kg]
        component mass
    platforms_mass : float, [kg]
        component mass
    transformer_mass : float, [kg]
        component mass
    tower_mass : float, [kg]
        component mass
    blade_number : float
        number of rotor blades
    main_bearing_number : float
        number of main bearings

    Returns
    -------
    hub_system_mass : float, [kg]
        hub system mass
    rotor_mass : float, [kg]
        hub system mass
    nacelle_mass : float, [kg]
        nacelle mass
    turbine_mass : float, [kg]
        turbine mass

    """

    def setup(self):
        # rotor
        self.add_input("blade_mass", 0.0, units="kg")
        self.add_input("hub_mass", 0.0, units="kg")
        self.add_input("pitch_system_mass", 0.0, units="kg")
        self.add_input("spinner_mass", 0.0, units="kg")

        # nacelle
        self.add_input("lss_mass", 0.0, units="kg")
        self.add_input("main_bearing_mass", 0.0, units="kg")
        self.add_input("gearbox_mass", 0.0, units="kg")
        self.add_input("hss_mass", 0.0, units="kg")
        self.add_input("brake_mass", 0.0, units="kg")
        self.add_input("generator_mass", 0.0, units="kg")
        self.add_input("bedplate_mass", 0.0, units="kg")
        self.add_input("yaw_mass", 0.0, units="kg")
        self.add_input("hvac_mass", 0.0, units="kg")
        self.add_input("cover_mass", 0.0, units="kg")
        self.add_input("platforms_mass", 0.0, units="kg")
        self.add_input("transformer_mass", 0.0, units="kg")

        # tower
        self.add_input("tower_mass", 0.0, units="kg")
        self.add_discrete_input("blade_number", 3)
        self.add_discrete_input("main_bearing_number", 2)

        self.add_output("hub_system_mass", 0.0, units="kg")
        self.add_output("rotor_mass", 0.0, units="kg")
        self.add_output("nacelle_mass", 0.0, units="kg")
        self.add_output("turbine_mass", 0.0, units="kg")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        blade_mass = inputs["blade_mass"]
        hub_mass = inputs["hub_mass"]
        pitch_system_mass = inputs["pitch_system_mass"]
        spinner_mass = inputs["spinner_mass"]
        lss_mass = inputs["lss_mass"]
        main_bearing_mass = inputs["main_bearing_mass"]
        gearbox_mass = inputs["gearbox_mass"]
        hss_mass = inputs["hss_mass"]
        brake_mass = inputs["brake_mass"]
        generator_mass = inputs["generator_mass"]
        bedplate_mass = inputs["bedplate_mass"]
        yaw_mass = inputs["yaw_mass"]
        hvac_mass = inputs["hvac_mass"]
        cover_mass = inputs["cover_mass"]
        platforms_mass = inputs["platforms_mass"]
        transformer_mass = inputs["transformer_mass"]
        tower_mass = inputs["tower_mass"]
        blade_number = discrete_inputs["blade_number"]
        bearing_number = discrete_inputs["main_bearing_number"]

        outputs["hub_system_mass"] = hub_mass + pitch_system_mass + spinner_mass
        outputs["rotor_mass"] = blade_mass * blade_number + outputs["hub_system_mass"]
        outputs["nacelle_mass"] = (
            lss_mass
            + bearing_number * main_bearing_mass
            + gearbox_mass
            + hss_mass
            + brake_mass
            + generator_mass
            + bedplate_mass
            + yaw_mass
            + hvac_mass
            + cover_mass
            + platforms_mass
            + transformer_mass
        )
        outputs["turbine_mass"] = outputs["rotor_mass"] + outputs["nacelle_mass"] + tower_mass


# --------------------------------------------------------------------


class nrel_csm_mass_2015(om.Group):
    def setup(self):
        self.add_subsystem("blade", BladeMass(), promotes=["*"])
        self.add_subsystem("hub", HubMass(), promotes=["*"])
        self.add_subsystem("pitch", PitchSystemMass(), promotes=["*"])
        self.add_subsystem("spinner", SpinnerMass(), promotes=["*"])
        self.add_subsystem("lss", LowSpeedShaftMass(), promotes=["*"])
        self.add_subsystem("bearing", BearingMass(), promotes=["*"])
        self.add_subsystem("torque", RotorTorque(), promotes=["*"])
        self.add_subsystem("gearbox", GearboxMass(), promotes=["*"])
        self.add_subsystem("hss", HighSpeedShaftMass(), promotes=["*"])
        self.add_subsystem("brake", BrakeMass(), promotes=["*"])
        self.add_subsystem("generator", GeneratorMass(), promotes=["*"])
        self.add_subsystem("bedplate", BedplateMass(), promotes=["*"])
        self.add_subsystem("yaw", YawSystemMass(), promotes=["*"])
        self.add_subsystem("hvac", HydraulicCoolingMass(), promotes=["*"])
        self.add_subsystem("cover", NacelleCoverMass(), promotes=["*"])
        self.add_subsystem("platforms", PlatformsMainframeMass(), promotes=["*"])
        self.add_subsystem("transformer", TransformerMass(), promotes=["*"])
        self.add_subsystem("tower", TowerMass(), promotes=["*"])
        self.add_subsystem("turbine", TurbineMassAdder(), promotes=["*"])


class nrel_csm_2015(om.Group):
    def setup(self):
        self.add_subsystem("nrel_csm_mass", nrel_csm_mass_2015(), promotes=["*"])
        self.add_subsystem("turbine_costs", Turbine_CostsSE_2015(verbosity=False), promotes=["*"])



import pandas as pd


class XlsxValidator:
    """
    XlsxValidator is for comparing the results of a previous model run
    to the results of a current model run.
    """

    def compare_expected_to_actual(self, expected_xlsx, actual_module_type_operation_list, validation_output_xlsx):
        """
        This compares the expected costs as calculated by a prior model run
        with the actual results from a current model run.

        It compares the results row by row and prints any differences.

        Parameters
        ----------
        expected_xlsx : str
            The absolute filename of the expected output .xlsx file.

        actual_module_type_operation_list : str
            The module_type_operation_list as returned by a subclass of
            XlsxManagerRunner.

        validation_output_xlsx : str
            The absolute pathname to the output file with the comparison
            results.

        Returns
        -------
        bool
            True if the expected and actual results are equal. It returns
            False otherwise.
        """
        # First, make the list of dictionaries into a dataframe, and drop
        # the raw_cost and raw_cost_total_or_per_turbine columns.
        actual_df = pd.DataFrame(actual_module_type_operation_list)
        actual_df.drop(["raw_cost", "raw_cost_total_or_per_turbine"], axis=1, inplace=True)
        expected_df = pd.read_excel(expected_xlsx, "costs_by_module_type_operation", engine="openpyxl")
        # expected_df = expected_df.dropna(inplace=True, how='all')
        expected_df.rename(
            columns={
                "Project ID with serial": "project_id_with_serial",
                "Number of turbines": "num_turbines",
                "Turbine rating MW": "turbine_rating_MW",
                "Module": "module",
                "Operation ID": "operation_id",
                "Type of cost": "type_of_cost",
                "Cost per turbine": "cost_per_turbine",
                "Cost per project": "cost_per_project",
                "USD/kW per project": "usd_per_kw_per_project",
            },
            inplace=True,
        )

        cost_per_project_actual = actual_df[
            ["cost_per_project", "project_id_with_serial", "module", "operation_id", "type_of_cost"]
        ]
        cost_per_project_expected = expected_df[
            ["cost_per_project", "project_id_with_serial", "module", "operation_id", "type_of_cost"]
        ]

        comparison = cost_per_project_actual.merge(
            cost_per_project_expected, on=["project_id_with_serial", "module", "operation_id", "type_of_cost"]
        )

        comparison.rename(
            columns={
                "cost_per_project_x": "cost_per_project_actual",
                "cost_per_project_y": "cost_per_project_expected",
            },
            inplace=True,
        )

        comparison["difference_validation"] = (
            comparison["cost_per_project_actual"] - comparison["cost_per_project_expected"]
        )

        # Regardless of the outcome, write the end result of the comparison
        # to the validation output file.
        columns_for_comparison_output = [
            "project_id_with_serial",
            "module",
            "operation_id",
            "type_of_cost",
            "cost_per_project_actual",
            "cost_per_project_expected",
            "difference_validation",
        ]
        comparison.to_excel(validation_output_xlsx, index=False, columns=columns_for_comparison_output)

        # If the comparison dataframe is empty, that means there are no common
        # projects in the expected data that match the actual data.
        if len(comparison) < 1:
            print("=" * 80)
            print("Validation error: There are no common projects between actual and expected data.")
            print("=" * 80)
            return False

        # Find all rows where the difference is unequal to 0. These are rows
        # that failed validation. Note that, after the join, the rows may be
        # in a different order than the originals.
        #
        # Round the difference to a given number of decimal places.
        failed_rows = comparison[comparison["difference_validation"].round(decimals=4) != 0]

        if len(failed_rows) > 0:
            print("=" * 80)
            print("The following rows failed validation:")
            print(failed_rows)
            print("=" * 80)
            return False
        else:
            return True
